from ursina import *



class PlatformerController2d(Entity):
    def __init__(self, **kwargs):
        super().__init__()

        self.model = 'cube'
        self.origin_y = -.5
        self.scale_y = 2
        self.color = color.orange
        self.collider = 'box'

        self.animator = Animator({'idle' : None, 'walk' : None, 'jump' : None})

        self.walk_speed = 8
        self.walking = False
        self.velocity = 0 # the walk direction is stored here. -1 for left and 1 for right.
        self.jump_height = 4
        self.jump_duration = .5
        self.jumping = False
        self.max_jumps = 1
        self.jumps_left = self.max_jumps
        self.gravity = 1
        self.grounded = True
        self.air_time = 0   # this increase while we're falling and used when calculating the distance we fall so we fall faster and faster instead of linearly.
        self.traverse_target = scene     # by default, it will collide with everything except itself. you can change this to change the boxcast traverse target.
        self.ignore_list = [self, ]
        self._start_fall_sequence = None # we need to store this so we can interrupt the fall call if we try to double jump.

        ray = raycast(self.world_position, self.down, distance=10, ignore=(self, ), traverse_target=self.traverse_target)
        if ray.hit:
            self.y = ray.world_point[1] + .01
        # camera.add_script(SmoothFollow(target=self, offset=[0,1,-30], speed=4))

        for key, value in kwargs.items():
            setattr(self, key, value)

        # delay_gravity one frame
        target_gravity = self.gravity
        self.gravity = 0
        invoke(setattr, self, 'gravity', target_gravity, delay=1/60)
        self._original_scale_x = self.scale_x

        self.min_x = -99999
        self.max_x = 99999

    # @every(1/30)
    # def fixed_update(self):
        # # if boxcast(self.world_position, direction=self.right, distance=self.scale_x/2)
        # hit_info = raycast(self.world_position+Vec3(0,self.scale_y/2,0), direction=self.right, distance=5, traverse_target=self.traverse_target, ignore=self.ignore_list, debug=True)
        # # hit_info_head = raycast(self.world_position+Vec3(0,self.scale_y*.99,0), direction=self.right, distance=abs(self.scale_x/2)+(self.walk_speed/20), traverse_target=self.traverse_target, ignore=self.ignore_list)
        # # if not self.grounded:
        # #     hit_info_feet = raycast(self.world_position+Vec3(0,.01,0), direction=self.right, distance=abs(self.scale_x/2)+(self.walk_speed/20), traverse_target=self.traverse_target, ignore=self.ignore_list)
        # # else:
        # #     from ursina.hit_info import HitInfo
        # #     hit_info_feet = HitInfo(hit=False)
        # # hit_infos = (hit_info, hit_info_head, hit_info_feet)
        # hit_infos = (hit_info, )
        # #
        # self.min_x = -99999
        # self.max_x = 99999
        # if any(hit_infos):
        #     if self.velocity < 0:
        #         self.min_x = max([hinf.world_point.x for hinf in hit_infos if hinf.hit])
        #     else:
        #         self.max_x = min([hinf.world_point.x for hinf in hit_infos if hinf.hit])


    def update(self):
        if not boxcast(
            self.position+Vec3(self.velocity * time.dt * self.walk_speed,self.scale_y/2,0),
            # self.position+Vec3(sefl,self.scale_y/2,0),
            direction=Vec3(self.velocity,0,0),
            distance=abs(self.scale_x/2),
            ignore=self.ignore_list,
            traverse_target=self.traverse_target,
            thickness=(self.scale_x*.99, self.scale_y*.9),
            ).hit:

            self.x += self.velocity * time.dt * self.walk_speed

        # self.x += self.velocity * time.dt * self.walk_speed
        # self.x = clamp(self.x, self.min_x-.5, self.max_x+.5)

        self.walking = held_keys['a'] + held_keys['d'] > 0 and self.grounded

        # animations
        if not self.grounded:
            self.animator.state = 'jump'
        else:
            if self.walking:
                self.animator.state = 'walk'
            else:
                self.animator.state = 'idle'


        # check if we're on the ground or not.
        ray = raycast(self.world_position+Vec3(0,.1,0), self.down, distance=max(.15, self.air_time * self.gravity), ignore=self.ignore_list, traverse_target=self.traverse_target)
        left_ray = raycast(self.world_position+Vec3(-self.scale_x*.49,.1,0), self.down, distance=max(.15, self.air_time * self.gravity), ignore=self.ignore_list, traverse_target=self.traverse_target)
        right_ray = raycast(self.world_position+Vec3(self.scale_x*.49,.1,0), self.down, distance=max(.15, self.air_time * self.gravity), ignore=self.ignore_list, traverse_target=self.traverse_target)

        # print(self.grounded)
        if any((ray.hit, left_ray.hit, right_ray.hit)):
            if not self.grounded:
                self.land()

            self.y = max(r.world_point.y for r in (ray, left_ray, right_ray) if r.hit)
            self.grounded = True
            return
        else:
            self.grounded = False

        # if not on ground and not on way up in jump, fall
        if not self.grounded and not self.jumping:
            # print(self.air_time)
            self.y -= min(self.air_time * self.gravity, ray.distance-.1)
            self.air_time += time.dt*4 * self.gravity   # fall faster and faster the long we've stayed in the air


        # if in jump and hit the ceiling, fall
        if self.jumping:
            # if boxcast(self.position+(0,.2,0), self.up, distance=self.scale_y, thickness=.95, ignore=self.ignore_list, traverse_target=self.traverse_target).hit:
            hit_above = raycast(self.world_position+Vec3(0,self.scale_y/2,0), self.up, distance=self.jump_height-(self.scale_y/2), traverse_target=self.traverse_target, ignore=self.ignore_list)
            hit_above_left = raycast(self.world_position+Vec3(-self.scale_x*.49,self.scale_y/2,0), self.up, distance=self.jump_height-(self.scale_y/2), traverse_target=self.traverse_target, ignore=self.ignore_list)
            hit_above_right = raycast(self.world_position+Vec3(self.scale_x*.49,self.scale_y/2,0), self.up, distance=self.jump_height-(self.scale_y/2), traverse_target=self.traverse_target, ignore=self.ignore_list)
            if any((hit_above.hit, hit_above_left.hit, hit_above_right.hit)):
                if hasattr(self, 'y_animator'):
                    self.y_animator.kill()
                self.air_time = 0
                self.start_fall()


    def input(self, key):
        if key == 'space':
            self.jump()

        if key == 'd':
            self.velocity = 1
            self.scale_x = self._original_scale_x
        if key == 'd up':
            self.velocity = -held_keys['a']

        if key == 'a':
            self.velocity = -1
        if key == 'a up':
            self.velocity = held_keys['d']

        if held_keys['d'] or held_keys['a']:
            self.scale_x = self._original_scale_x * self.velocity


    def jump(self):
        if not self.grounded and self.jumps_left <= 1:
            return

        if self._start_fall_sequence:
            self._start_fall_sequence.kill()

        if hasattr(self, 'y_animator'):
            self.y_animator.kill()
        self.jump_dust = Entity(model=Circle(), scale=.5, color=color.white33, position=self.position)
        self.jump_dust.animate_scale(3, duration=.3, curve=curve.linear)
        self.jump_dust.fade_out(duration=.2)
        destroy(self.jump_dust, 2.1)

        self.jumping = True
        self.jumps_left -= 1
        self.grounded = False

        target_y = self.y + self.jump_height
        duration = self.jump_duration

        self.animate_y(target_y, duration, resolution=30, curve=curve.out_expo)
        self._start_fall_sequence = invoke(self.start_fall, delay=duration)


    def start_fall(self):
        if hasattr(self, 'y_animator'):
            self.y_animator.pause()
        self.jumping = False


    def land(self):
        # print('land')
        self.air_time = 0
        self.jumps_left = self.max_jumps
        self.grounded = True



if __name__ == '__main__':
    # window.vsync = False
    app = Ursina()
    camera.orthographic = True
    camera.fov = 10

    ground = Entity(model='cube', color=color.white33, origin_y=.5, scale=(20, 1, 1), collider='box', y=-1)
    wall = Entity(model='cube', color=color.azure, origin=(-.5,.5), scale=(5,10), x=10, y=.5, collider='box')
    wall_2 = Entity(model='cube', color=color.white33, origin=(-.5,.5), scale=(5,10), x=10, y=0, collider='box')
    ceiling = Entity(model='cube', color=color.white33, origin_y=-.5, scale=(1, 1, 1), y=1, collider='box')
    ceiling = Entity(model='cube', color=color.white33, origin_y=-.5, scale=(5, 5, 1), y=2, collider='box')
    ground = Entity(model='cube', color=color.white33, origin_y=.5, scale=(20, 3, 1), collider='box', y=-1, rotation_z=45, x=-5)

    def input(key):
        if key == 'c':
            wall.collision = not wall.collision
            print(wall.collision)


    player_controller = PlatformerController2d(scale_y=2, jump_height=4, x=3, y=20, max_jumps=2)
    ec = EditorCamera()
    ec.add_script(SmoothFollow(target=player_controller, offset=[0,1,0], speed=4))

    app.run()



from ursina import *

app = Ursina()
points = LoopingList([Draggable(parent=scene, model='circle', scale=.25, x=-4+(i*1)) for i in range(6)])

mover = Entity(model='quad', scale=.3, t=0, color=color.red, position=points[0].position)
mover_2 = Entity(model='quad', scale=1, t=0, color=color.lime, speed=0)

line_renderer = Entity(model=Mesh(vertices=[Vec3(0), Vec3(0)], mode='line', thickness=5), z=.1)
def update():
    line_renderer.model.vertices = [e.position for e in points]
    line_renderer.model.generate()
    mover_2.look_at_2d(mover)
    # mover_2.position += mover_2.up * mover_2.speed * time.dt
    mover_2.position = lerp(mover_2.position, mover.position, time.dt*2)

speed = 10
def play():
    mover.position = points[0].position
    mover_2.position = mover.position

    delay = 0
    for i, e in enumerate(points):
        if i == len(points)-1:
            break

        dist_to_next = distance_2d(points[i].position, points[i+1].position)

        # invoke(setattr, mover, 'position', points[i].position, delay=delay/speed)
        mover.animate_position(points[i+1].position, duration=dist_to_next/speed, delay=delay/speed, curve=curve.linear)
        delay += dist_to_next


play_button = Button(text='play', y=-.3, color=color.azure, on_click=play)
play_button.fit_to_text()


app.run()



from ursina.shader import Shader

camera_contrast_shader = Shader(fragment='''
#version 430

uniform sampler2D tex;
uniform float contrast = 1.;

in vec2 uv;
out vec4 out_color;



void main() {
    vec4 color = texture(tex, uv).rgba;
    color.rgb = clamp(mix(vec3(0.5, 0.5, 0.5), color.rgb, contrast), 0., 1.);   // adjust contrast
    out_color = color;
}


''',
default_input={
    'contrast': 1,
}
)




if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='sphere')
    e = Entity(model='cube', y=-1)
    camera.shader = camera_contrast_shader
    camera.set_shader_input('contrast', 1)

    slider = ThinSlider(max=2, dynamic=True, position=(-.25, -.45))
    def adjust_contrast():
        camera.set_shader_input("contrast", slider.value)
    slider.on_value_changed = adjust_contrast

    EditorCamera()

    app.run()



from ursina import *; ssao_shader = Shader(fragment=''' #version 430
uniform sampler2D tex;
uniform sampler2D dtex;
uniform sampler2D ntex;
in vec2 uv;

uniform sampler2D noise_texture;
in flat int numsamples;
in float radius;
in float amount;
in float strength;
in float falloff;

uniform vec4 params1;
uniform vec4 params2;

out vec4 color;


uniform vec2 InverseViewport;
// uniform vec4 SolidColor;

// uniform sampler2D SourceImage;
// uniform sampler2D DepthImage;
// uniform sampler2D NormalsImage;
uniform sampler2D PositionsImage;
// uniform sampler2D NoiseImage;

// uniform float Time = 0;

uniform vec2 TexelSize;
uniform float OccluderBias = 0.375;
uniform float SamplingRadius = 11.0;
uniform vec2 Attenuation = vec2(1,1);

uniform bool ApplyVignette = true;

out vec4 FragColor;

// Mostly cribbed from http://devmaster.net/posts/3095/shader-effects-screen-space-ambient-occlusion
float SamplePixels (vec3 srcPosition, vec3 srcNormal, vec2 uv)
{
    // Get the 3D position of the destination pixel
    // vec3 dstPosition = texture(PositionsImage, uv).xyz;
    vec3 dstPosition = vec3(uv.x, uv.y, texture(dtex, uv).r);

    // Calculate ambient occlusion amount between these two points
    // It is simular to diffuse lighting. Objects directly above the fragment cast
    // the hardest shadow and objects closer to the horizon have minimal effect.
    vec3 positionVec = dstPosition - srcPosition;

    float bias = OccluderBias;
    float z = texture(dtex, uv).w;
    bias += z * 0.5;

    float intensity = max(dot(normalize(positionVec), srcNormal) - bias, 0.0);

    // Attenuate the occlusion, similar to how you attenuate a light source.
    // The further the distance between points, the less effect AO has on the fragment.
    float dist;
    dist = length(positionVec) / SamplingRadius;
    float attenuation = 1.0 / (Attenuation.x + (Attenuation.y * dist));

    return intensity * attenuation;
}

// Mostly cribbed from http://devmaster.net/posts/3095/shader-effects-screen-space-ambient-occlusion
float ComputeOcclusion(vec2 tc, vec3 srcNormal)
{
    // Get position and normal vector for this fragment
    // vec3 srcPosition = texture(PositionsImage, tc).xyz;
    vec3 srcPosition = vec3(uv.x, uv.y, texture(dtex, uv).r);
    vec2 randVec = normalize(texture(noise_texture, tc*20).xy * 2.0 - 1.0);
    // float srcDepth = texture(PositionsImage, tc).w;
    float srcDepth = texture(dtex, tc).r;

    // The following variable specifies how many pixels we skip over after each
    // iteration in the ambient occlusion loop. We can't sample every pixel within
    // the sphere of influence because that's too slow. We only need to sample
    // some random pixels nearby to apprxomate the solution.
    //
    // Pixels far off in the distance will not sample as many pixels as those close up.
    float kernelRadius = SamplingRadius * (1.0 - srcDepth);

    // Sample neighbouring pixels
    vec2 kernel[4];
    kernel[0] = vec2(0.0, 1.0); // top
    kernel[1] = vec2(1.0, 0.0); // right
    kernel[2] = vec2(0.0, -1.0);    // bottom
    kernel[3] = vec2(-1.0, 0.0);    // left

    const float Sin45 = 0.707107;   // 45 degrees = sin(PI / 4)

    // Sample from 16 pixels, which should be enough to appromixate a result. You can
    // sample from more pixels, but it comes at the cost of performance.
    float occlusion = 0.0;
    for (int i = 0; i < 4; ++i) {
        vec2 k1 = reflect(kernel[i], randVec);
        vec2 k2 = vec2(k1.x * Sin45 - k1.y * Sin45,
                k1.x * Sin45 + k1.y * Sin45);
        k1 *= TexelSize;
        k2 *= TexelSize;

        occlusion += SamplePixels(srcPosition, srcNormal, tc + k1 * kernelRadius);
        occlusion += SamplePixels(srcPosition, srcNormal, tc + k2 * kernelRadius * 0.75);
        occlusion += SamplePixels(srcPosition, srcNormal, tc + k1 * kernelRadius * 0.5);
        occlusion += SamplePixels(srcPosition, srcNormal, tc + k2 * kernelRadius * 0.25);
    }

    // Average and clamp ambient occlusion
    occlusion /= 16.0;
    return clamp(occlusion, 0.0, 1.0);
}

void main() {
    // vec2 tc = uv * InverseViewport;
    vec2 tc = uv;

    vec4 c = texture(tex, tc);

    bool bkgd = (c.a == 0);
    // c = mix(c,vec4(1),0.55); // brighten up for pastel-like colors

    vec3 n = texture(ntex, uv).rgb;

    // The geometry has inconsistent normals...
    // So flip 'em if necessary
    // if (dot(n,vec3(0,0,1)) < 0)
    //     n = -n;

    float ov = 1 - ComputeOcclusion(tc, n) *10;
    ov = ov*ov*ov;
    FragColor = ov*c;

    if (bkgd) {
        FragColor = vec4(0.627, 0.322, 0.176, 1);
    }

    // if (ApplyVignette) {
    //     float d = distance(tc, vec2(0.5));
    //     FragColor.rgb *= 1 - 0.75 * d;
    // }
}
''',
default_input = dict(
    TexelSize = 32,
    OccluderBias = 0.375,
    SamplingRadius = 11.0,
    Attenuation = Vec2(1,1),

)
)
app = Ursina()

camera.clip_plane_far = 100
camera.shader = ssao_shader
camera.set_shader_input('noise_texture', load_texture('noise')._texture)


Entity(model='plane', color=color.gray, scale=8)
Entity(model='cube', color=color.gray, y=.5)
Entity(model='sphere', color=color.azure, y=1, x=.75)

# from ursina.shaders import colored_lights_shader
# for e in scene.entities:
#     if not e.eternal:
#         e.shader = colored_lights_shader

base.render.setShaderAuto()

EditorCamera()
app.run()



import shutil
import subprocess

from ursina import *


class VideoRecorder(Entity):
    def __init__(self, max_duration=5, fps=30, name='untitled_video', **kwargs):
        super().__init__()
        self.recording = False
        self.file_path = Path(application.asset_folder) / 'video_temp'
        self.i = 0
        self.max_duration = max_duration
        self.fps = fps
        self.video_name = name
        self.t = 0

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.max_frames = int(self.max_duration * self.fps)


    def start_recording(self):
        print('start recording,', self.max_duration, self.file_path)
        window.fps_counter.enabled = False
        window.exit_button.visible = False
        self.max_frames = self.max_duration * self.fps

        if self.file_path.exists():
            shutil.rmtree(self.file_path)   # delete temp folder

        self.file_path.mkdir(parents=True, exist_ok=True)
        # base.movie(namePrefix=f'\\video_temp\\{self.video_name}', max_duration=2.0, fps=30, format='mp4', sd=4)

        application.calculate_dt = False
        time.dt = 1 / self.fps
        self.recording = True
        # invoke(self.stop_recording, delay=self.max_duration)



    def stop_recording(self):
        self.recording = False
        window.fps_counter.enabled = True
        window.exit_button.visible = True
        print('stop recording')
        # self.convert_to_gif()
        # command = 'ffmpeg -framerate 60 -f image2 -i video_temp/%04d.png -c:v libvpx-vp9 -pix_fmt yuva420p untitled_video.webm'
        command = (
            f'ffmpeg -y '
            f'-framerate {self.fps} '
            f'-i "{self.file_path}/{self.video_name}_%04d.png" '
            f'-c:v libx264 -crf 18 -pix_fmt yuv420p '
            f'"{self.video_name}.mp4"'
        )
        subprocess.Popen(command, shell=True)
        application.calculate_dt = True

        # print('converting to video:', result)
        print('saved webm to:', Path(f'{self.file_path.parent}/{self.video_name}.webm'))

    def update(self):
        if not self.recording:
            return

        time.dt = 1/self.fps
        base.screenshot(namePrefix=self.file_path / f'{self.video_name}_{str(self.i).zfill(4)}.png', defaultFilename=0)
        self.i += 1
        if self.i >= self.max_frames:
            self.stop_recording()


    def input(self, key):
        combo = ['control', 'r', 'e', 'c']
        if key in combo and all([held_keys[e] for e in combo]):
            if not self.recording:
                self.start_recording()
            else:
                self.stop_recording()


class VideoRecorderUI(WindowPanel):
    def __init__(self, **kwargs):
        self.duration_label = Text('duration:')
        self.duration_field = InputField(default_value='5')
        self.fps_label = Text('fps:')
        self.fps_field = InputField(default_value='30')
        self.name_label = Text('name:')
        self.name_field = InputField(default_value='untitled_video')

        self.start_button = Button(text='Start Recording [Shift+F12]', color=color.azure, on_click=self.start_recording)

        super().__init__(
            title='Video Recorder [F12]',
            content=(
                self.duration_label,
                self.duration_field,
                self.fps_label,
                self.fps_field,
                self.name_label,
                self.name_field,
                Space(1),
                self.start_button,
                ),
            )
        self.y = .5
        self.scale *= .75
        self.visible = False


    def input(self, key):
        if key == 'f12':
            self.visible = not self.visible

        if held_keys['shift'] and key == 'f12':
            self.start_button.on_click()


    def start_recording(self):
        print(self.name_field)
        if self.name_field.text == '':
            self.name_field.blink(color.hsv(0,1,1,.5), .5)
            print('enter name')
            return

        # self.start_button.color=color.lime
        self.visible = False
        application.video_recorder.max_duration = float(self.max_duration.text)
        application.video_recorder.video_name = self.name_field.text
        application.video_recorder.frame_skip = 60 // int(self.fps_field.text)
        application.video_recorder.recording = True



if __name__ == '__main__':
    app = Ursina()
    window.size = (1280*.5, 720*.5)
    from ursina.prefabs.first_person_controller import FirstPersonController
    from ursina.shaders import lit_with_shadows_shader
    random.seed(0)
    Entity.default_shader = lit_with_shadows_shader

    ground = Entity(model='plane', collider='box', scale=64, texture='grass', texture_scale=(4,4))

    editor_camera = EditorCamera(enabled=False, ignore_paused=True)
    player = FirstPersonController(model='cube', z=-10, color=color.orange, origin_y=-.5, speed=8)
    player.collider = BoxCollider(player, Vec3(0,1,0), Vec3(1,2,1))

    gun = Entity(model='cube', parent=camera, position=(.5,-.25,.25), scale=(.3,.2,1), origin_z=-.5, color=color.red, on_cooldown=False)

    shootables_parent = Entity()
    mouse.traverse_target = shootables_parent

    for i in range(16):
        Entity(model='cube', origin_y=-.5, scale=2, texture='brick', texture_scale=(1,2),
            x=random.uniform(-8,8),
            z=random.uniform(-8,8) + 8,
            collider='box',
            scale_y = random.uniform(2,3),
            color=color.hsv(0, 0, random.uniform(.9, 1))
            )


    sun = DirectionalLight()
    sun.look_at(Vec3(1,-1,-1))
    Sky()

    vr = VideoRecorder(max_duration=120)



    app.run()



from ursina import *

class ReplaceTextureButton():

    def __init__(self):
        self.path = None

    def input(self, key):
        if key == 'left mouse down' and self.entity.hovered:
            if self.path:
                self.replace_texture()

    @undoable
    def replace_texture(self):
        self.target_entities = list()
        self.original_textures = list()
        for s in scene.editor.selection:
            self.original_textures.append(str(s.texture.getFullpath()))
            self.target_entities.append(s)

        for s in scene.editor.selection:
            s.texture = self.path

        scene.editor.inspector.update_inspector()

        # undo
        yield 'replace texture'
        print('--------------', self.original_textures)
        for i in range(len(self.target_entities)):
            self.target_entities[i].texture = self.original_textures[i]
        scene.editor.inspector.update_inspector()



from ursina.audio import Audio
from ursina.ursinastuff import invoke
from ursina import curve
from ursina.string_utilities import print_warning
from ursina.ursinastuff import after
from typing import Literal



tracks = dict()
current_music_track = ''
current_ambiance_track = ''


def _load_audio(track_name, audio_group='music'):
    audio_instance = Audio(track_name, loop=True, autoplay=False, group=audio_group, ignore_paused=True)
    if not audio_instance.clip:
        print_warning('track not found:', track_name)
        return None
    tracks[track_name] = audio_instance
    return audio_instance


def play(track_name, fade_out_duration=2, start=0, track_group:Literal['music','ambiance']='music'):
    global current_music_track, current_ambiance_track

    if track_group == 'music':
        current_track = current_music_track
    elif track_group == 'ambiance':
        current_track = current_ambiance_track
    else:
        print_warning(f'Invalid audio group: {track_group}')
        return

    if track_name == current_track:
        return

    if not track_name and current_track:
        tracks[current_track].fade_out(fade_out_duration)
        if track_group == 'music':
            current_music_track = None
        elif track_group == 'ambiance':
            current_ambiance_track = None
        return

    if track_name not in tracks:
        audio_group = track_group
        if track_group == 'ambiance':
            audio_group = 'ambient'
        audio = _load_audio(track_name, audio_group)
        if audio is None:
            print('audio not found')
            return

    if not current_track:  # if no music/ambiance playing, start immediately
        tracks[track_name].play(start)
        if track_group == 'music':
            current_music_track = track_name
        elif track_group == 'ambiance':
            current_ambiance_track = track_name
        return

    # fade out current track and play new one after
    print(f'{track_group}_system: fade out:', current_track)
    tracks[current_track].fade_out(duration=fade_out_duration, curve=curve.linear, destroy_on_ended=False, ignore_paused=True)
    print(f'{track_group}_system: fade in:', track_name)

    @after(fade_out_duration, ignore_paused=True)
    def _():
        tracks[track_name].play(start=start)
        tracks[track_name].fade_in(duration=fade_out_duration)
    if track_group == 'music':
        current_music_track = track_name
    elif track_group == 'ambiance':
        current_ambiance_track = track_name


def play_ambiance(track_name, fade_out_duration=2, start=0):
    play(track_name, fade_out_duration, start, track_group='ambiance')


if __name__ == '__main__':
    from ursina import *
    from ursina import music_system

    app = Ursina()

    # music_system.play('noise', track_group='music')
    # music_system.play('forest', track_group='ambiance')

    # def input(key):
    #     if key == 'space':
    #         if music_system.current_music_track == 'noise':
    #             music_system.play('square', track_group='music')
    #         else:
    #             music_system.play('noise', track_group='music')

    #     if key == 'a':
    #         if music_system.current_ambiance_track == 'forest':
    #             music_system.play('rain', track_group='ambiance')
    #         else:
    #             music_system.play('forest', track_group='ambiance')
    music_changer = ButtonGroup(('None', 'crestlands_part', 'dunes_part'), label='music', )
    def on_music_selected():
        music_system.play(music_changer.value if music_changer.value != 'None' else None)
    music_changer.on_value_changed = on_music_selected

    ambiance_changer = ButtonGroup(('None', 'noise', 'square'), label='ambiance', y=-.1)
    def on_ambiance_selected():
        music_system.play_ambiance(ambiance_changer.value if ambiance_changer.value != 'None' else None)
    ambiance_changer.on_value_changed = on_ambiance_selected

    t = Text('SPACE: Toggle music\nA: Toggle ambiance', origin=(0, 0), y=-.4)
    def update():
        t.text = f'''\
            current_music_track:    {music_system.current_music_track},
            current_ambiance_track: {music_system.current_ambiance_track},
            '''

    app.run()



fragment_code = fragment='''
#version 420
uniform float iTime;
in vec2 iResolution;

in vec2 fragCoord;
out vec4 fragColor;

void main()
{
    vec2 uv = fragCoord;
    vec3 col = 0.5 + 0.5*cos(uv.xyx+vec3(0,4,8));
    fragColor = vec4(col,1.0);
}
'''
from panda3d.core import Shader

camera_contrast_shader = Shader.make(Shader.SL_GLSL,
vertex='''
#version 420
uniform mat4 p3d_ModelViewProjectionMatrix;
in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
out vec2 fragCoord;


void main() {
  gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
  fragCoord = p3d_MultiTexCoord0;
}
''',

fragment=fragment_code, geometry='')


if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='quad', scale=3)
    e.shader = camera_contrast_shader

    t = 0
    frame = 0

    e.set_shader_input('iResolution', window.size)
    e.set_shader_input('iTime', t)
    e.set_shader_input('iFrame', frame)

    def update():
      global t, frame
      t += time.dt
      e.set_shader_input('iTime', t)

      frame += 1
      e.set_shader_input('iFrame', frame)


    EditorCamera()

    app.run()



from math import sqrt


def distance(a, b):
    return sqrt((b[0] - a[0])**2 + (b[1] - a[1])**2 + (b[2] - a[2])**2)


def merge_overlapping_vertices(vertices, triangles=None, max_distance=.1):

    if triangles == None:
        triangles = [(i, i+1, i+2) for i in range(0, len(vertices), 3)]

    if triangles[0] is int:
        triangles = [(triangles[i], triangles[i+1], triangles[i+2]) for i in range(0, len(triangles), 3)]

    unique = list()
    triangles = list(triangles)

    for i, v in enumerate(vertices):
        for j, u in enumerate(unique):
            if distance(v, u) <= max_distance:
                # print('found overlapping vertices!', i, 'could be', j)
                for k, t in enumerate(triangles):
                    if t == i:
                        triangles[k] = j
                        # for v in verts

        unique.append(v)

    return unique, triangles


if __name__ == '__main__':
    verts = ((0,0,0), (1,0,0), (1,1,0), (0,0,0), (1,1,0), (0,1,0))
    tris = (0,1,2,3,4,5)

    new_verts, new_tris = merge_overlapping_vertices(verts, tris)
    print('verts:', (verts), (new_verts))
    print('tris:', (tris), (new_tris))

    from ursina import *
    app = Ursina()

    e = Entity(model=Mesh(new_verts, new_tris, mode='triangle'))
    EditorCamera()

    app.run()
    # m = Cylinder(8)
    # calculate_normals(m.vertices)



fragment_code = fragment='''
#version 130

out vec4 fragColor;
uniform sampler2D p3d_Texture0;
uniform int osg_FrameNumber;
in vec3 cam_pos;
in float textureWeight;
in vec3 initialRay;

float intersectSphere(vec3 origin, vec3 ray, vec3 center, float redius) {
    vec3 toSphere = origin - center;
    float a = dot(ray, ray);
    float b = 2.0 * dot(toSphere, ray);
    float c = dot(toSphere, toSphere) - redius*redius;
    float discriminant = b*b - 4.0*a*c;

    if(discriminant > 0.0) {
        float t = (-b - sqrt(discriminant)) / (2.0 * a);
        if(t > 0.0)
            return t;
        }
    return 10000.0;
}
float random(vec3 scale, float seed) {
  return fract(sin(dot(gl_FragCoord.xyz + seed, scale)) * 43758.5453 + seed);
}

vec3 calculateColor(vec3 origin, vec3 ray, vec3 light) {
    // if (random(vec3(.001,.1,.001), osg_FrameNumber) > 0.5) {
    //     return vec3(1,1,1);
    //
    // }
    vec3 colorMask = vec3(1.0);
    vec3 accumulatedColor = vec3(1,0,0);

    for(int bounce = 0; bounce < 5; bounce++) {
    // int bounce = 0;
    // vec2 tRoom = intersectCube(origin, ray, roomCubeMin, roomCubeMax);
        vec2 tRoom = vec2(0,0);
        float tSphere0 = intersectSphere(origin, ray, vec3(0,0,0), .25);
        vec3 surfaceColor = vec3(0.8, 0.8, 0.8);
        float t = 10000.0;

        if(tRoom.x < tRoom.y)
            t = tRoom.y;

        if(tSphere0 < t) {
            t = tSphere0;
            surfaceColor = vec3(0.0, 1.0, 0.1);
        }

        vec3 hit = origin + ray * t;

        float specularHighlight = 0.0;
        vec3 normal;

        if(t == tRoom.y) {
            normal = -normalForCube(hit, roomCubeMin, roomCubeMax);
            if(hit.x < -4.99895)
                surfaceColor = vec3(0.1, 0.1, 0.1);
            else if(hit.x > 4.99895)
                surfaceColor = vec3(0.1, 0.1, 0.1);
                ray = cosineWeightedDirection(timeSinceStart + float(bounce), normal);
        }

        else if(t == 10000.0) {
        if(t == 10000.0) {
            break;
        }
        else {
            if(false);

            // else if(t == tSphere0)
            //     normal = normalForSphere(hit, vec3(0,0,0), .25);

            // ray = cosineWeightedDirection(timeSinceStart + float(bounce), normal);
        }
    vec3 toLight = light - hit;
    float diffuse = max(0.0, dot(normalize(toLight), normal));
    // float shadowIntensity = shadow(hit + normal * 0.0001, toLight);
    colorMask *= surfaceColor;
    float shadowIntensity = 1;
    accumulatedColor += colorMask * (0.5 * diffuse * shadowIntensity);
    accumulatedColor += colorMask * specularHighlight * shadowIntensity;
    origin = hit;
    }
    return accumulatedColor;
}

void main() {
    fragColor = vec4(0,1,0,1);
    vec3 texture = texture2D(p3d_Texture0, gl_FragCoord.xy / vec2(1920, 1080)).rgb;

    vec3 newLight = vec3(1,1,1); //temp
    // int textureWeight = 1;

    fragColor = vec4(mix(calculateColor(cam_pos, initialRay, newLight), texture, textureWeight), 1.0);
}
'''









from panda3d.core import Shader

camera_contrast_shader = Shader.make(Shader.SL_GLSL,
vertex='''
#version 130


uniform mat4 p3d_ModelViewProjectionMatrix;
in vec4 p3d_Vertex;
// in vec2 p3d_MultiTexCoord0;
// out vec2 texcoord;
uniform vec3 cam_pos, ray00, ray01, ray10, ray11;
out vec3 initialRay;

void main() {
    vec2 percent = p3d_Vertex.xy * 0.5 + 0.5;
    initialRay = mix(mix(ray00, ray01, percent.y), mix(ray10, ray11, percent.y), percent.x);
    // gl_Position = p3d_Vertex;
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
    // texcoord = p3d_MultiTexCoord0;
}
''',

fragment=fragment_code, geometry='')


if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='cube', color=color.orange)
    camera.shader = camera_contrast_shader
    # camera.set_shader_input('contrast', 1)

    cam_pos = Vec3(0,0,0)

    t = 0

    mat = camera.lens.getProjectionMat()

    camera.set_shader_input('cam_pos', cam_pos)
    # camera.set_shader_input('light', Vec3(0.4, 0.5, -0.6))
    camera.set_shader_input('ray00', mat[0][0])
    camera.set_shader_input('ray01', mat[0][1])
    camera.set_shader_input('ray10', mat[1][0])
    camera.set_shader_input('ray11', mat[1][1])

    camera.set_shader_input('sphereCenter0', Vec3(0,1,1))
    camera.set_shader_input('sphereRadius0', .25)

    sampleCount = 0
    textureWeight = sampleCount / (sampleCount + 1)

    def update():
      global t, sampleCount
      t += 1
      #t += time.dt * .5
      #t = t % 1
      # // print(t)

      sampleCount += 1
      textureWeight = sampleCount / (sampleCount + 1)
      camera.set_shader_input('sampleCount', sampleCount)
      camera.set_shader_input('textureWeight', textureWeight)


    EditorCamera()
    window.size *= .5
    window.position += Vec2(900, -100)

    app.run()



Nacho Gouda #cook
    worker


Pierre Inkwell
# writer from west pardam
# writes artisitc and political poetry
    worker, drink lover, cat lover, artistic



Halibut Hotel
# a hotel for the middle class
# mostly empty due to low tourism
    if not player.home == self:
        "Hi there player, are you looking to stay at the Halibut tonight?"
            "Yes, as a matter of fact I do. Do you have any rooms left?"
        if missions.hotel_advertising.completed:
            "No we don't, and it's all thanks to you!"
            "Here, have this as a thank you..."
            player.inventory.append(items.fancy_watch)
        else:
            "Yes, would you like to stay here for the night?"
                "Yes, please"
                    player.gold -= 20
                "Maybe another time."
                exit()
    else:
        "I hope you didn't lose your key."
            "No, of course not. whisper: ...actually I did lose it."

    if player is Teri:
        "What do you want?"

shop
    a room for the night    # hide from the police



from ursina import *


app = Ursina()

r = 8
for i in range(1, r):
    t = i/r
    s = 4*i
    print(s)
    grid = Entity(model=Grid(s,s), scale=s, color=color.hsv(0,0,.8,lerp(.8,0,t)), rotation_x=90, y=i/1000)
    subgrid = duplicate(grid)
    subgrid.model = Grid(s*4, s*4)
    subgrid.color = color.hsv(0,0,.4,lerp(.8,0,t))
    EditorCamera()

app.run()



import sys
sys.path.append("..")
from ursina import *
import os
import uuid
from collections import defaultdict
from itertools import chain
import time

class HierarchyPanel(Entity):

    def __init__(self):
        super().__init__()
        self.name = 'hierarchy_panel'
        self.parent = scene.ui
        self.is_editor = True
        # self.model = 'quad'
        # self.color = color.panda_button
        self.size = 1
        print('scene:', scene)
        self.origin = (-.5, .5)
        self.position = window.top_left
        self.scale = (.249, 1)
        self.editor_collider = 'box'

        self.button_parent = Entity()
        self.button_parent.parent = self
        self.button_parent.y = -.025 * self.size
        self.button_parent.z = -1
        self.max_vertical = 1000
        self.button_size = (1, .025 * self.size)

        self.bg = Entity()
        self.bg.model = 'quad'
        self.bg.color = color.panda_button
        self.bg.parent = self
        self.bg.origin = (-.5, .5)

        self.temp_hierarchy_panel = list()
        self.buttons = list()
        self.bg.add_script('scrollable')
        self.bg.scrollable.target = self.button_parent
        self.t = 0
        self.i = 0

        self.populate()


    def populate(self):

        self.i = 0
        for child in self.button_parent.children:
            destroy(child)
        self.buttons.clear()

        print('-------pop----------')
        for e in scene.children:
            self.traverse_tree(e)

        self.bg.y = (- (self.i + 1) * (self.button_size[1] + .001)) + .001


    def create_button(self, entity, name):
        button = EditorButton()
        button.is_editor = True
        button.parent = self.button_parent
        button.origin = (-.5, .5)
        button.position = (
            0,
            (-self.i * (self.button_size[1] + .001)))
        button.scale = self.button_size * scene.editor_size
        button.color = color.panda_button
        button.text = name[0:min(32, len(name))]
        # button.text = name
        button.text_entity.align = 'left'
        button.text_entity.x = -.5
        # button.text_entity.wordwrap = button.model.getScale(render)[0] * 4

        selection_button = button.add_script('selection_button')
        selection_button.selection_target = entity
        self.buttons.append(button)

        self.i += 1

        return button


    def traverse_tree(self, e, indent=''):
        indent += '  '
        self.create_button(e, indent + e.name)
        for c in e.children:
            self.traverse_tree(c, indent)

        # print(self.buttons)


    def input(self, key):
        if key == 'left mouse down' and self.hovered:
            scene.editor.selection.clear()
            for b in self.buttons:
                b.color = color.panda_button



from ursina.editor.level_editor import *
from ursina.shaders import colored_lights_shader
from ursina.scripts.property_generator import generate_properties_for_class

@generate_properties_for_class()
class PokeShape(Entity):
    default_values = Entity.default_values | dict(
        name='poke_shape',
        # make_wall=True,
        wall_height=1.0,
        # wall_thickness=.1,
        subdivisions=0,
        smoothing_distance=.1,
        points=[Vec3(-.5,0,-.5), Vec3(.5,0,-.5), Vec3(.5,0,.5), Vec3(-.5,0,.5)],
        collider_type='None',
        texture='grass',
        texture_scale=Vec2(.125, .125),
        # shader_inputs={'side_texture':Func(load_texture, 'grass'), }

    ) # combine dicts

    gizmo_color = color.violet

    def __init__(self, edit_mode=False, **kwargs):
        kwargs = __class__.default_values | kwargs
        points = kwargs.pop('points', None)
        self.ready = False
        super().__init__(**kwargs)

        self.original_parent = LEVEL_EDITOR
        self.selectable = True
        self.highlight_color = color.blue
        # self._point_gizmos = LoopingList([Entity(parent=self, original_parent=self, position=e, selectable=False, name='PokeShape_point', is_gizmo=True) for e in kwargs['points']])
        self.model = Mesh()
        self.add_new_point_renderer = Entity(model=Mesh(mode='point', vertices=[], thickness=.075), color=color.white, alpha=.5, texture='circle', unlit=True, is_gizmo=True, selectable=False, enabled=False, always_on_top=True)
        self.add_collider = False
        self._wall_parent = None
        self.wall_height = kwargs['wall_height']
        self.subdivisions =  kwargs['subdivisions']
        self.smoothing_distance = kwargs['smoothing_distance']

        self._point_gizmos = []
        if not points:
            self.points = __class__.default_values['points']
        else:
            self.points = points

        self.texture = kwargs['texture']

        self.position = kwargs['position']
        for key in Entity.default_values.keys():
            if key == 'model':
                continue
            setattr(self, key, kwargs[key])

        # if edit_mode:
        self.edit_mode = edit_mode

        self.generate()
        self.ready = True


    def draw_inspector(self):
        return {'edit_mode': bool, 'wall_height': float, 'subdivisions':int, 'smoothing_distance':float}


    def generate(self):
        # print('--------------', self.texture_scale)
        import tripy
        # if not self.model:
        #     return
        self._point_gizmos = LoopingList([e for e in self._point_gizmos if e])   # ensure deleted points are removed
        polygon = LoopingList(Vec2(*e.get_position(relative_to=self).xz) for e in self._point_gizmos)

        if self.subdivisions:
            for j in range(self.subdivisions):
                smooth_polygon = LoopingList()
                for i, p in enumerate(polygon):
                    smooth_polygon.append(lerp(p, polygon[i-1], self.smoothing_distance))
                    smooth_polygon.append(lerp(p, polygon[i+1], self.smoothing_distance))
                polygon = smooth_polygon

        triangles = tripy.earclip(polygon)
        self.model.vertices = []
        for tri in triangles:
            for v in tri:
                self.model.vertices.append(Vec3(v[0], 0, v[1]))

        self.model.uvs = [Vec2(v[0],v[2])*1 for v in self.model.vertices]
        self.model.normals = [Vec3(0,1,0) for i in range(len(self.model.vertices))]
        self.model.generate()
        # self.texture = 'grass'
        # [destroy(e) for e in self.wall_parent.children]
        # print('-------------', self.make_wall, self.wall_parent)
        if self._wall_parent:
            # print('destroy old wall parent')
            destroy(self._wall_parent)
            self._wall_parent = None

        if self.wall_height:
            if not self._wall_parent:
                # print('make new wall parent')
                self._wall_parent = Entity(parent=self, model=Mesh(), color=color.dark_gray, add_to_scene_entities=False, shader=colored_lights_shader)

            # polygon_3d = [Vec3(e[0], 0, e[1]) for e in polygon]
            # polygon_3d.append(polygon_3d[0])
            # self.wall_parent.model = Pipe(base_shape=Quad(scale=(self.wall_thickness, self.wall_height)), path=polygon_3d)
            # self.wall_parent.model = Pipe(polygon_3d, path=[Vec3(0,0,0), Vec3(0,-10,0)])
            wall_verts = []
            for i, vert in enumerate(polygon):
                vert = Vec3(vert[0], 0, vert[1])
                next_vert = Vec3(polygon[i+1][0], 0, polygon[i+1][1])

                wall_verts.extend((
                    vert,
                    vert + Vec3(0,-self.wall_height,0),
                    next_vert,

                    next_vert,
                    vert + Vec3(0,-self.wall_height,0),
                    next_vert + Vec3(0,-self.wall_height,0),
                ))
            #     # wall = Entity(model='cube', origin_x=-.5, scale=.1, position=vert, scale_x=distance(vert, next_vert), color=color.blue, parent=self._wall_parent, add_to_scene_entities=False)
            #     # wall.look_at(next_vert, 'right')
            #
            self._wall_parent.model.vertices = wall_verts
            self._wall_parent.model.generate_normals(False)
            self._wall_parent.model.generate()


        # if self.add_collider:
        #     self.collider = self.model

        if self.edit_mode:
            self.add_new_point_renderer.model.vertices = []
            for i, e in enumerate(self._point_gizmos):
                self.add_new_point_renderer.model.vertices.append(lerp(self._point_gizmos[i].world_position, self._point_gizmos[i+1].world_position, .5))
                # self.add_new_point_renderer.model.vertices.append(self._point_gizmos[i].world_position)
            self.add_new_point_renderer.model.generate()


    def __deepcopy__(self, memo):
        changes = self.get_changes(__class__)

        _copy = __class__(texture_scale = self.texture_scale, **changes)
        _copy.texture_scale = self.texture_scale
        # print('---------------', _copy.texture_scale)
        return _copy


    def points_getter(self):
        return [e.position for e in self._point_gizmos]

    def points_setter(self, value):
        [destroy(e) for e in self._point_gizmos]
        self._point_gizmos = LoopingList([Entity(parent=self, original_parent=self, position=e, selectable=False, name='PokeShape_point', is_gizmo=True, enabled=False) for e in value])
        LEVEL_EDITOR.entities.extend(self._point_gizmos)


    def edit_mode_getter(self):
        return getattr(self, '_edit_mode', False)

    def edit_mode_setter(self, value):
        self._edit_mode = value

        print('set edit mode', value)
        if value:
            [setattr(e, 'selectable', False) for e in LEVEL_EDITOR.entities if not e == self]
            for e in self._point_gizmos:
                if not e in LEVEL_EDITOR.entities:
                    LEVEL_EDITOR.entities.append(e)

            [setattr(e, 'selectable', True) for e in self._point_gizmos]
            LEVEL_EDITOR.gizmo.subgizmos['y'].enabled = False
            LEVEL_EDITOR.gizmo.fake_gizmo.subgizmos['y'].enabled = False
            self.add_new_point_renderer.enabled = True
            self.collider = None
        else:
            [LEVEL_EDITOR.entities.remove(e) for e in self._point_gizmos if e in LEVEL_EDITOR.entities]
            [setattr(e, 'selectable', True) for e in LEVEL_EDITOR.entities]
            if True in [e in LEVEL_EDITOR.selection for e in self._point_gizmos]: # if point is selected when exiting edit mode, select the poke shape
                LEVEL_EDITOR.selection = [self, ]

            LEVEL_EDITOR.gizmo.subgizmos['y'].enabled = True
            LEVEL_EDITOR.gizmo.fake_gizmo.subgizmos['y'].enabled = True
            self.add_new_point_renderer.enabled = False
            self.collider = 'mesh'
        LEVEL_EDITOR.render_selection()

    # @property
    # def wall_height(self):
    #     return self._wall_height
    # @wall_height.setter
    # def wall_height(self, value):
    #     self._wall_height = value
    #     self.generate()

    def update(self):
        if self.edit_mode:
            if mouse.left or held_keys['d']:
                LEVEL_EDITOR.render_selection()
                self.generate()


    def input(self, key):
        combined_key = input_handler.get_combined_key(key)
        if combined_key == 'tab':
            if not LEVEL_EDITOR.selection:
                self.edit_mode = False

            if self in LEVEL_EDITOR.selection or True in [e in LEVEL_EDITOR.selection for e in self._point_gizmos]:
                self.edit_mode = not self.edit_mode

        if self.edit_mode and (key == 'left mouse down' or key == 'd'):
            if LEVEL_EDITOR.selector.get_hovered_entity():
                return
            points_in_range = [(distance_2d(world_position_to_screen_position(v), mouse.position), v) for v in self.add_new_point_renderer.model.vertices]
            points_in_range = [e for e in points_in_range if e[0] < .075/2]
            points_in_range.sort()

            closest_point = None
            if not points_in_range:
                return

            closest_point = points_in_range[0][1]
            i = self.add_new_point_renderer.model.vertices.index(closest_point)

            new_point = Entity(parent=self, original_parent=self, position=lerp(self._point_gizmos[i].position, self._point_gizmos[i+1].position, .5), selectable=True, is_gizmo=True)
            LEVEL_EDITOR.entities.append(new_point)
            self._point_gizmos.insert(i+1, new_point)
            LEVEL_EDITOR.render_selection()
            if key == 'd':
                LEVEL_EDITOR.quick_grabber.input('d')


        elif key == 'space':
            self.generate()

        # elif key == 'double click' and LEVEL_EDITOR.selector.get_hovered_entity() == self:
        #     self.edit_mode = True
        #
        # elif key == 'double click' and not LEVEL_EDITOR.selector.get_hovered_entity():
        #     self.edit_mode = False

        elif self.edit_mode and key.endswith(' up'):
            invoke(self.generate, delay=3/60)

    # def __setattr__(self, name, value):
    #     if name == 'model' and hasattr(self, 'model') and self.model and not isinstance(value, Mesh):
    #         print_info('can\'t set model of PokeShape')
    #         return
    #
    #     super().__setattr__(name, value)


if __name__ == '__main__':
    app = Ursina(borderless=False)

    level_editor = LevelEditor()
    level_editor.goto_scene(0,0)
    # cube = WhiteCube(selectable=True)
    # level_editor.entities.append(PokeShape())
    level_editor.add_entity(PokeShape(points=[Vec3(-6.89023, 0, -5.93539), Vec3(-5.63213, 0, -6.7236), Vec3(-3.06749, 0, -7.45143), Vec3(0.883525, 0, -6.64059), Vec3(6.21342, 0.000496293, -5.19114), Vec3(11.1816, 0.000748294, -1.60608), Vec3(13.0414, 0.000874309, 0.223267), Vec3(12.6511, 0.0010004, 2.84322), Vec3(9.07899, 0.000750429, 5.98706), Vec3(5.59802, 0.000500321, 5.69713), Vec3(3.45835, 0.000375334, 7.03647), Vec3(2.95372, 0.000250343, 9.16615), Vec3(4.31376, 0.00012526, 9.91672), Vec3(5.6031, 0, 12.5358), Vec3(4.99113, 0.00049994, 13.8873), Vec3(3.37031, 0.000749853, 15.6645), Vec3(-0.513243, 0.000874871, 16.5425), Vec3(-2.08884, 0.000999762, 15.3409), Vec3(-3.86994, 0.000500111, 16.2354), Vec3(-5.61374, 0.000375315, 18.7785), Vec3(-5.64462, 0.000250529, 22.5197), Vec3(-18.9718, 0.000125618, 15.9756), Vec3(-14.6812, 0, 10.564), Vec3(-14.2555, 0, 7.90265), Vec3(-13.5866, 0, 4.08694), Vec3(-10.7991, 0, 1.16432), Vec3(-9.05981, 0, 1.75484), Vec3(-7.52061, 0, 0.920164), Vec3(-6.02536, 0, -1.79266), Vec3(-7.2474, 0, -3.23652)]))
    app.run()



from ursina import *
app  = Ursina()

m = model=Mesh(vertices=load_model('quad', use_deepcopy=True).vertices, mode='point', thickness=10)
e = Entity()
# e.setShaderAuto()
e.model = m
EditorCamera()
Entity(model='plane', color=color.black, scale=10)
print(e.shader)
app.run()


from ursina import *

app = Ursina()


Entity(model='plane', scale=10, texture='brick')
EditorCamera()

camera.y = 5
camera.look_at((0,0,0))

# filter
from panda3d.core import Shader
from direct.filter.FilterManager import FilterManager
from panda3d.core import Texture as PandaTexture

# manager = FilterManager(base.win, camera.ui_camera)
manager = FilterManager(base.win, base.cam)
tex = PandaTexture()
quad = manager.renderSceneInto(colortex=tex)
quad.setShader(Shader.load("projector_shader.sha"))
quad.setShaderInput("tex", tex)

quad.setShaderInput("projector_texture", load_texture('projector_texture')._texture)



app.run()



from ursina import *
from ursina.prefabs.first_person_controller import FirstPersonController

application.asset_folder = application.asset_folder.parent
app = Ursina()

Entity(
    model='mysterious_sphere',
    texture='mysterious_sphere_ao_map',
    collider='mesh'
    )

Sky()
player = FirstPersonController()
player.cursor.enabled = False
player.speed *= .5
window.exit_button.visible = False
# window.fps_counter.enabled = False

# Entity(model='plane', scale=1000, color=color.azure)
sun = Entity(y=10, z=1000)

Entity(model='sphere', y=2, z=-8, color=color.red, scale=3)


# from direct.filter.CommonFilters import CommonFilters
# filters = CommonFilters(base.win, base.cam)
#
# filters.setAmbientOcclusion()
# filters.setVolumetricLighting(sun)


app.run()



from ursina import *
from copy import copy, deepcopy
from pprint import pprint



color.lime = color.color(60, .9, .77)
color.azure = color.color(180, .56, .73)
app = Ursina()
ground_plane = Entity(model='plane', scale=32, origin=(-.5,0,-.5), position=(-.5,0,-.5), collider='box', color=color.azure, visible=False)
water = Entity(model='plane', scale=256, position=(16,-1,16), color=color.azure)
cursor = Entity(model='cube', origin=(-.5,-.5,-.5), scale_y=1, color=color.color(0,1,.8,.3))
fake_ground = Entity(model='plane', scale=36, position=(-.5,.01,-.5), color=color.azure.tint(-.2), origin=(-.5,0,-.5))

fallback = Mesh(
    vertices=[(-.5,0,-.5), (.5,0,-.5), (.5,0,.5), (-.5,0,.5),  (-.5,0,-.5), (.5,0,-.5), (.5,0,.5), (-.5,0,.5), (-.5,-16,-.5), (.5,-16,-.5), (.5,-16,.5), (-.5,-16,.5)],
    triangles=[(0,1,2,3), (4,5,6,7), (4,8,9,5), (5,9,10,6), (6,10,11,7), (7,11,8,4)],
    colors=(color.lime, )*4 + (color.dark_gray, )*8
    )

middle = Mesh(vertices=((-.5,0,-.5), (.5,0,-.5), (.5,0,.5), (-.5,0,.5)), triangles=((0,1,2,3),), colors=(color.lime, )*4)

edge = Mesh(
    vertices=((-.5,0,-.5), (.5,0,-.5), (.5,0,.5), (-.5,0,.5),       (-.5,0,.5), (.5,0,.5), (.5,-1,.5), (-.5,-1,.5)),
    triangles=((0,1,2,3), (4,5,6,7)),
    colors=(color.lime, )*2 + (color.lime.tint(.1), )*2 + (color.dark_gray, )*4
    )

outer_corner = Mesh(
    vertices=((-.5,0,-.5), (.5,0,-.5), (-.5,0,.5),      (-.5,0,.5), (.5,0,-.5), (.5,-1,-.5), (-.5,-1,.5),     (-.5,-1,.5), (.5,-1,-.5), (.5,-1,.5)),
    triangles=((0,1,2), (3,4,5,6), (7,8,9)),
    colors=(color.lime, color.lime.tint(.1), color.lime.tint(.1)) + (color.dark_gray, )*4 +(color.lime, )*3
    )

inner_corner = Mesh(
    vertices=[(-.5,0,-.5), (.5,0,-.5), (.5,0,.5), (-.5,0,.5)],
    triangles=[(0,1,2,3), ],
    colors=(color.lime, color.lime, color.lime.tint(.1), color.lime)
    )
# edge = Entity(model='quad').model

# corner_shape_name = deepcopy(cube)
# corner_shape_name.c
valid_corners = dict()
for e in ('00001110', '00001111', '00011110', '00011111'):
    for i in range(4):
        e = e[i*2:] + e[:i*2]   # rotate
        # valid_corners[e] = i*90
        rotations = (0, -90, 90, 180)
        valid_corners[e] = rotations[i]

valid_edges = dict()
for e in ('00111110', '01111110', '00111111', '01111111'):
    for i in range(4):
        e = e[i*2:] + e[:i*2]   # rotate
        # rotations = (-180, 90, -90, 0)
        rotations = (0, -90, 90, 180)
        valid_edges[e] = rotations[i]



world_size = 32
entities = [[Entity(model=copy(middle), position=(x,0,z), ignore=True, scale_y=1, shape_name='middle') for z in range(world_size)] for x in range(world_size)]
brush_size = 1
target_height = 4
x, z = 0, 0

def get_heights():
    return [[e.y for e in row] for row in entities]

undo_cache = list()
undo_cache.append(get_heights())
undo_index = len(undo_cache)-1

def update():
    global x, y, target_height

    if mouse.hovered_entity == ground_plane:
        cursor.world_position = mouse.world_point
        cursor.x = round(cursor.x)
        cursor.z = round(cursor.z)
        x, z = math.floor(cursor.x), math.floor(cursor.z)
        x, z = clamp(x, 0, world_size-1), clamp(z, 0, world_size-1)
        # cursor.y = entities[x][z].y

        # sample height
        if held_keys['alt'] and mouse.left:
            target_height = entities[x][z].y

        # draw
        elif mouse.left:
            for j in range(brush_size):
                for i in range(brush_size):
                    temp_x, temp_z = clamp(x+i, 0, world_size-1), clamp(z+j, 0, world_size-1)
                    entities[temp_x][temp_z].y = target_height

            find_shape()


def find_shape():
    for z in range(1,world_size-1):
        for x in range(1, world_size-1):
            # check neighbors
            target_height = entities[x][z].y
            n, ne, e, se, s, sw, w, nw = True, True, True, True, True, True, True, True
            n =  entities[x  ][z+1].y >= target_height
            ne = entities[x+1][z+1].y >= target_height
            e =  entities[x+1][z  ].y >= target_height
            se = entities[x+1][z-1].y >= target_height
            s =  entities[x  ][z-1].y >= target_height
            sw = entities[x-1][z-1].y >= target_height
            w =  entities[x-1][z  ].y >= target_height
            nw = entities[x-1][z+1].y >= target_height

            neighbors = ''.join([str(int(e)) for e in (n, ne, e, se, s, sw, w, nw)])
            e = entities[x][z]

            if not '0' in neighbors:    # middle
                if e.shape_name != 'middle':
                    e.shape_name = 'middle'
                    e.model = copy(middle)
                    # e.origin_y=.5

                    # e.color = color.white

            elif neighbors.count('0') == 1 and neighbors.index('0') % 2 == 1:   # inner corner
                if e.shape_name != 'inner_corner':
                    e.shape_name = 'inner_corner'
                    e.model = copy(inner_corner)
                    # e.color = color.orange
                    e.rotation_y = (neighbors.index('0')-1) / 2 * 90

            elif neighbors in valid_edges:    # edge
                if e.shape_name != 'edge':
                    e.shape_name = 'edge'
                    e.model = copy(edge)
                    # e.origin_y=.5
                    e.rotation_y = valid_edges[neighbors]

                    pos = e.position + e.forward
                    bottom_neighbor_height = entities[int(pos[0])][int(pos[2])].y
                    bottom_neighbor_height = e.y - bottom_neighbor_height
                    # Entity(model='sphere', color=color.red, position=e.position + e.forward, y=bottom_neighbor_height)
                    m = eval(edge.recipe)
                    # print('-------------', bottom_neighbor_height, e.y)
                    m.vertices = [Vec3(v[0], -bottom_neighbor_height, v[2]) if v[1] < -.5 else v for v in m.vertices]
                    m.generate()
                    e.model = m
                # e.color = color.lime

            elif neighbors in valid_corners:    # outer corner
                if e.shape_name != 'outer_corner':
                    e.shape_name = 'outer_corner'
                    e.rotation_y = valid_corners[neighbors]
                    # print('-----', valid_corners[neighbors])
                    # print(e.position + e.forward)
                    pos = e.position + e.forward
                    bottom_neighbor_height = entities[int(pos[0])][int(pos[2])].y
                    bottom_neighbor_height = e.y - bottom_neighbor_height
                    # Entity(model='sphere', color=color.red, position=e.position + e.forward, y=bottom_neighbor_height)
                    m = eval(outer_corner.recipe)
                    # print('-------------', bottom_neighbor_height, e.y)
                    #
                    new_verts = list()
                    for v in m.vertices:
                        if v[1] < -.5:
                            new_verts.append(Vec3(v[0], -bottom_neighbor_height, v[2]))
                            continue

                        new_verts.append(v)

                    m.vertices = new_verts
                    m.generate()
                    e.model = m

                    # e.model = copy(outer_corner)
                    # e.origin_y=.5
                    # e.rotation_y = valid_corners[neighbors]
                # e.color = color.red

            else:
                if e.shape_name != 'fallback':
                    e.shape_name = 'fallback'
                    e.model = copy(fallback)


def save(name):
    # print(entities)
    for e in entities[0]:
        print(e.y, )


def load(name):
    # Autoblock()
    pass


def input(key):
    global undo_cache, undo_index, entities, brush_size, target_height, cursor
    if not mouse.left:
        if held_keys['control']:
            if key in ('z', 'z hold') and not held_keys['shift']:
                undo_index -= 1
                undo_index = clamp(undo_index, 0, len(undo_cache)-1)
                for z in range(world_size):
                    for x in range(world_size):
                        entities[x][z].y = undo_cache[undo_index][x][z]

            if held_keys['shift'] and key in ('z', 'z hold') or key in ('y', 'y hold'):
                undo_index += 1
                undo_index = clamp(undo_index, 0, len(undo_cache)-1)
                for z in range(world_size):
                    for x in range(world_size):
                        entities[x][z].y = undo_cache[undo_index][x][z]


        if mouse.hovered_entity == ground_plane and key == 'left mouse up':
            # record undo
            undo_index += 1
            undo_cache = undo_cache[:undo_index]
            undo_cache.append(get_heights())

    if key == 'd':
        brush_size += 1
        brush_size = max(brush_size, 1)
        cursor.scale = (brush_size,1,brush_size)
    if key == 'x':
        brush_size -= 1
        brush_size = max(brush_size, 1)
        cursor.scale = (brush_size,1,brush_size)

    if key == 'q':
        target_height -= 1
        target_height = max(target_height, 0)
    if key == 'e':
        target_height += 1
        target_height = max(target_height, 0)

    if held_keys['control'] and key == 's':
        self.save()



EditorCamera(position=(16,10,16), rotate_around_mouse_hit=False)
# camera.orthographic = True
window.exit_button.visible = False
window.fps_counter.enabled = False
window.color = color.black
app.run()



from ursina import *
from ursina.shaders import unlit_shader


class Sky(Entity):
    default_values = dict(parent=camera, name='sky', model='sky_dome', texture='sky_default', scale=9900, shader=unlit_shader, unlit=True)
    instances = []

    def __init__(self, **kwargs):
        super().__init__(**(__class__.default_values | kwargs))

        # self.setDepthWrite(False)
        __class__.instances.append(self)
        self.setBin('background', 0)


    def update(self):
        self.world_rotation = Vec3(0,0,0)
        self.scale = camera.clip_plane_far * .8

if __name__  == '__main__':
    app = Ursina()
    Sky(texture='sky_sunset')
    camera.fov = 90
    EditorCamera()

    # test
    def input(key):
        if key == '-':
            camera.clip_plane_far -= 100 +  (held_keys['control']*10)
            print(camera.clip_plane_far)
        elif key == '+':
            camera.clip_plane_far += 100 + (held_keys['control']*10)
            print(camera.clip_plane_far)

    app.run()



from ursina import *




# app = Ursina()
#
#
# Entity(model='cube', scale=8, collider='box')
#
#
# sphere = Entity(model='sphere', scale=10, collider='sphere')
#
# EditorCamera()
#
#
# def input(key):
#     if key == 'left mouse down' and mouse.point:
#         e = Entity(model='cube', scale=.5, color=color.gray, position=mouse.world_point)
#         # print(mouse.world_normal, mouse.collision)
#         e.look_at(e.position + mouse.world_normal)
#
#     if key == 'f':
#         sphere.model.colorize()
#         for i, v in enumerate(sphere.model.vertices):
#             # print(v)
#             # e = Entity(parent=sphere.model, model='cube', world_scale=.5, color=color.gray, position=v, add_to_scene_entities=False)
#             # e.look_at(e.position + sphere.model.normals[i])
#             # e.world_parent = scene
#             # sphere.model.colors[i] = color.random_color()
#             sphere.model.colors[i] = color.black
#
#             ray = raycast(v + sphere.model.normals[i], sphere.model.normals[i], debug=True)
#             if not ray.hit:
#                 sphere.model.colors[i] = color.white
#
#
#
#         sphere.model.generate()
#
#
app = Ursina()

ray = Entity()
def hit_sphere(center, radius, ray):
    oc = ray.position - center
    a = dot(r.forward, r.forward)
    b = 2 * dot(oc, r.forward)


t = Entity(parent=camera.ui, model='quad', texture='brick').texture
t.filtering = False
print(t)

def render():
    i = time.time()

    lower_left_corner = Vec3(-2, -1, -1)
    horizontal = Vec3(4, 0, 0)
    vertical = Vec3(0, 2, 0)
    origin = Vec3(0, 0, 0)

    for y in range(t.height):
        for x in range(t.width):
            col = lerp(color.red, color.yellow, x/t.width)
            col = lerp(col, color.azure, (y/t.height))
            t.set_pixel(x, y, col)

    t.apply()

    print(time.time() - i)


render()






app.run()



import pyximport
import os
import sys
import inspect
import importlib.util
import subprocess
import sysconfig
from pathlib import Path

# Install pyximport hooks for automatic compilation
pyximport.install()


def get_function_code(func):
    code_obj = func.__code__        # Extract the code object from the function
    # Get the function's source code (this assumes you have access to the source code)
    try:
        # Retrieve the actual source code of the function (if available)
        import inspect
        source_code = inspect.getsource(func)
        if source_code[0] == '@':
            source_code = 'def ' + source_code.split('\ndef ', 1)[1]    # remove decorators
            print('--...................\n', source_code)
    except Exception as e:
        print(f'Error retrieving function source: {e}')
        return None

    return source_code


def list_variables_and_types(source_code, annotations):
    import re
    # Use regex to find assignments in the source code
    pattern = r"(\w+)\s*[:=]\s*(.*)"
    variables = re.findall(pattern, source_code)
    
    variable_types = {}

    # Loop over the variables to try to infer types from the annotations
    # print('-------', func.__annotations__)
    for varname, expr in variables:
        if varname in variable_types:
            continue

        var_annotation = annotations.get(varname, None)
        if var_annotation is not None:
            variable_types[varname] = var_annotation
        try:
            # Evaluate the expression to guess the type
            # print('fffffffffffffff', expr)
            # if not expr.startswith('(') and not expr.startswith('[') and '(' in expr:
            #     variable_types[varname] = expr.split('(',1)[0]
            #     continue
            if expr.startswith('Vec3('):
                variable_types[varname] = Vec3
                continue
                

            evaluated_value = eval(expr)
            # print('aa', varname, evaluated_value)
            variable_types[varname] = type(evaluated_value)
        except Exception as e:
            # If evaluation fails, mark type as Unknown
            # print('e:', e)
            variable_types[varname] = None

    return source_code, variable_types


def guess_types(source_code, annotations):
    source_code, variables = list_variables_and_types(source_code, annotations)
    # print('aaaaaaaaaaaaaaaaaaaaaadding types to ', source_code, variables)
    for name, vartype in variables.items():
        if not vartype:
            continue
        # print(':::::::::::::::::::::', name, ':', vartype.__name__)
        if vartype == Vec3:
            # print('REPLACPEL VEC###################################', name, f' = {vartype.__name__}(' in source_code)
            source_code = source_code.replace(f' = {vartype.__name__}(', f' = (')
            source_code = source_code.replace(f'{name} = ', f'{name}:{'tuple[cython.double, cython.double, cython.double]'} = ')
            # print('REPLACPEL VEC###################################', source_code)
        # fubcc
    source_code = source_code.replace(':Vec3,', ':tuple[cython.double, cython.double, cython.double],')
    source_code = source_code.replace(':Vec3)', ':tuple[cython.double, cython.double, cython.double])')
    source_code = source_code.replace(' -> Vec3:', ' -> tuple[cython.double, cython.double, cython.double]:')

    return source_code


def compile_pyx_file(pyx_file):
    # Get the name of the compiled shared object
    c_file_path = pyx_file.with_suffix('').with_suffix('.c')
    compiled_extention_module = pyx_file.with_suffix('').with_suffix('.so')
    python_interpreter = sys.executable
    include_path = sysconfig.get_paths()['include']

    subprocess.run([
        python_interpreter, 
        '-m', 
        'cython', 
        '-a', # annotated 
        str(pyx_file), 
        '-o', 
        str(c_file_path)
    ], check=True)

    subprocess.run([
        'gcc',
        '-shared',
        '-fPIC', 
        '-O3',
        f'-I{include_path}',
        str(c_file_path), 
        "-o", 
        str(compiled_extention_module)
    ], check=True)
    
    return compiled_extention_module



def pyximport(pyx_file, module_name):
    compiled_extention_module = compile_pyx_file(pyx_file)
    
    # Add the path to the compiled file to sys.path
    module_dir = compiled_extention_module.parent
    if module_dir not in sys.path:
        sys.path.append(module_dir)
    
    # Dynamically import the compiled module
    spec = importlib.util.spec_from_file_location(module_name, compiled_extention_module)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    
    return module


def load_module_from_path(module_file_path, name=''):
    if not name:
        name = module_file_path.stem

    if not module_file_path.exists():
        raise FileNotFoundError(f'The module file {module_file_path} does not exist.')

    if module_file_path.suffix == '.pyx':
        pyximport(module_file_path, name)

    elif module_file_path.suffix == '.py':
        sys.path.append(str(module_file_path.parent))
    
    else:
        raise ValueError('Only .py and .pyx files are supported.')

    module = importlib.import_module(name)
    return module


__autocythonize__ = True

from functools import wraps

def cythonize_function(func=None, include=None):
    '''
    A decorator to compile a Python function into a Cython function at runtime.
    '''
    if include is None:
        include = [] 
    else:
        original_include_count = len(include)
        include = tuple(set(include))
        if len(include) < original_include_count:
            print_warning('Included more than one copy in:', func)


    if func is None:
        return lambda func: cythonize_function(func, include=include)
    # def decorator(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not __autocythonize__:
            return func(*args, **kwargs)

        # Get the current script's filename without extension
        script_name = Path(__file__).stem
        func_name = func.__name__
        current_file = Path(__file__)

        # Construct a unique filename based on the script name and the function name
        pyx_file_path = current_file.parent / '__cythonized__' / f'{current_file.stem}_{func_name}.pyx'
        # print('---------', pyx_file_path)
        import textwrap
        module_code = textwrap.dedent('''\
            # generated by cythonize_function decorator
            cimport cython

            ''')

        print("Functions to include:", include)
        for incl in include:
            include_code = get_function_code(incl)
            annotations = incl.__annotations__
            include_code = guess_types(include_code, annotations)
            # print('------------------iiiiiiiiiiiiiiiiii', 'including function:', incl.__name__, include_code)
            module_code += textwrap.dedent('''\
                @cython.boundscheck(False)  # Disable bounds checking for performance
                @cython.wraparound(False)   # Disable negative indexing
                @cython.cfunc               # C function, not called from Python
            ''') + include_code + '\n\n'
            # module_code += include_code

        func_code = get_function_code(func)
        annotations = func.__annotations__
        func_code = guess_types(func_code, annotations)

        module_code += textwrap.dedent('''\
            @cython.boundscheck(False)  # Disable bounds checking for performance
            @cython.wraparound(False)   # Disable negative indexing
        ''') + func_code

        pyx_file_path.parent.mkdir(exist_ok=True)

        with open(pyx_file_path, 'w') as pyx_file:
            pyx_file.write(module_code)

        # Ensure the current directory is in the sys.path
        sys.path.insert(0, pyx_file_path.parent)
        # print('paths:', sys.path)

        import importlib
        # print('importing:', str(pyx_file_path), 'as', pyx_file_path.stem)

        compiled_module = load_module_from_path(pyx_file_path, pyx_file_path.stem)
        # print('compiled module:', compiled_module)

        # Replace the original function with the compiled one
        cython_func = getattr(compiled_module, func_name)
        sys.path.pop(0)
        return cython_func(*args, **kwargs)
        # os.unlink(pyx_file_path)    # Clean up the temporary file

    return wrapper
    # return decorator

# __autocythonize__ = False
# Example usage
import cython

from ursina import Vec3

def _subtract(v1:Vec3, v2:Vec3) -> Vec3:
    result = Vec3(v1[0] - v2[0], v1[1] - v2[1], v1[2] - v2[2])
    return result

@cythonize_function(include=[_subtract, ])
def example_function():
    # from cythonizer__subtract import _subtract
    compiled = 'compiled' if __file__.endswith('.so') else 'pure python'
    from time import perf_counter
    import random
    t = perf_counter()
    
    result = Vec3(0.0, 0.0, 0.0)

    # result: tuple_with_three_floats = (0.0, 0.0, 0.0)
    other = Vec3(1.0, 2.0, 3.0)
    for i in range(1_000_000):
        other = Vec3(1.0, 2.0, 3.0)
        result = _subtract(result, other)

    print('-----:', compiled, perf_counter() - t)
    return result

# Test the function
if __name__ == '__main__':
    ...
    print(_subtract(Vec3(2,2,2), Vec3(1,1,1)))
    result = example_function()
    print('Result:', result)



import sys

from ursina import application
from ursina.scene import instance as scene
from ursina.ursinastuff import invoke
from ursina.sequence import Sequence


def destroy(entity, delay=0, unscaled=True, ignore_paused=False):
    if application.development_mode:
        # get the calling function and the file it's from, so we can give a better error message if we try to use it after destroy
        entity.destroy_source = f'caller: {sys._getframe(1).f_code.co_name} file: {sys._getframe(1).f_code.co_filename}'

    if delay == 0:
        _destroy(entity)
        return True

    return invoke(_destroy, entity, delay=delay, unscaled=unscaled, ignore_paused=ignore_paused)
    # return Sequence(Wait(delay), Func(_destroy, entity), auto_destroy=True, started=True)


def _destroy(entity, force_destroy=False):
    # from ursina import camera
    # if not entity or entity == camera:
    #     return

    if entity.eternal and not force_destroy:
        return

    if hasattr(entity, 'scripts'):
        for s in entity.scripts:
            del s

    if hasattr(entity, 'animations'):
        for anim in entity.animations:
            anim.kill()

    for child in entity.children:
        _destroy(child)

    if entity.collider:
        entity.collider.remove()

    if hasattr(entity, 'clip') and hasattr(entity, 'stop'): # stop audio
        entity.stop(False)

    if hasattr(entity, 'on_destroy'):
        entity.on_destroy()

    if entity in scene.entities:
        scene._entities_marked_for_removal.append(entity)

    if entity in scene.collidables:
        scene.collidables.remove(entity)

    if hasattr(entity, '_parent') and entity._parent and hasattr(entity._parent, '_children') and entity in entity._parent._children:
        entity._parent._children.remove(entity)

    for e in entity.loose_children:
        destroy(e)

    if hasattr(entity, '_loose_parent') and entity._loose_parent and hasattr(entity._loose_parent, '_loose_children') and entity in entity._loose_parent._loose_children:
        entity._loose_parent._loose_children.remove(entity)

    if hasattr(entity, 'tooltip'):
        _destroy(entity.tooltip)

    if hasattr(entity, '_on_click') and isinstance(entity._on_click, Sequence):
        entity._on_click.kill()

    if entity.hasPythonTag("Entity"):
        entity.clearPythonTag("Entity")

    entity.removeNode()

    if hasattr(entity.__class__, 'instances') and entity in entity.__class__.instances:
        entity.__class__.instances.remove(entity)
    #unload texture
    # if hasattr(entity, 'texture') and entity.texture != None:
    #     entity.texture.releaseAll()

    del entity



if __name__ == '__main__':
    from ursina import Entity, Ursina
    app = Ursina()
    class E(Entity):
        def __init__(self, name):
            super().__init__()
            self.num_frames = 0
            self.name = name

        def update(self):
            self.num_frames += 1
            print(f"updating {self}")
            if self.name == "e2" and self.num_frames == 3:
                print("destroying e2")
                destroy(self)


    app = Ursina(window_type="none")
    e1 = E("e1")
    e2 = E("e2")
    e3 = E("e3")


    def update():
        print()
        if e1.num_frames > 4:
            exit()


    app.run()


from pathlib import Path
from panda3d.core import Shader as Panda3dShader
from ursina import application

default_vertex_shader = '''
#version 430
uniform mat4 p3d_ModelViewProjectionMatrix;
in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
out vec2 uv;

void main() {
  gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
  uv = p3d_MultiTexCoord0;
}
'''

default_fragment_shader='''
#version 430

uniform sampler2D tex;
in vec2 uv;
out vec4 color;

void main() {
    vec3 rgb = texture(tex, uv).rgb;
    color = vec4(rgb, 1.0);
}

'''
imported_shaders = dict()

def do_shader_includes(shader_source, included=None):
    if shader_source is None:
        return None
    if shader_source == "":
        return ""
    inc = None
    if included is not None:
        inc = included
    else:
        inc = set()
    lines = shader_source.split("\n")
    res = ""
    for line in lines:
        if line.startswith("#include"):
            str_start = line.find("\"")
            include_str = line[str_start:]
            str_end = include_str[1:].find("\"")
            if str_end == -1:
                raise Exception("Missing closing quotation mark in GLSL include.")
            include_path = include_str[1:-1]
            include_path = include_path.replace("ursina/", str(application.package_folder / "shaders") + "/")
            if include_path not in inc:
                inc.add(include_path)
                with open(include_path, "r") as f:
                    file_contents = f.read() + "\n"
                    file_contents = do_shader_includes(file_contents, included=inc)
                    res += file_contents
        else:
            res += line + "\n"
    return res

class Shader:
    CG = Panda3dShader.SL_Cg
    GLSL = Panda3dShader.SL_GLSL
    HLSL = Panda3dShader.SL_HLSL
    SPIR_V = Panda3dShader.SL_SPIR_V

    def __init__(self, name='untitled_shader', language=GLSL, vertex=default_vertex_shader, fragment=default_fragment_shader, geometry='', **kwargs):

        if not ('__compiled__' in globals() and str(globals()['__compiled__'].__class__.__name__) == '__nuitka_version__'):
            from inspect import getframeinfo, stack
            _stack = stack()
            _caller = getframeinfo(_stack[1][0])
            self.path = Path(_caller.filename)

        self.name = name
        self.language = language
        self.vertex = vertex
        self.fragment = fragment
        self.geometry = geometry

        self.default_input = dict()
        self.continuous_input = dict()
        self.compiled = False
        if self not in imported_shaders:
            imported_shaders[self.name] = self

        for key, value in kwargs.items():
            setattr(self, key ,value)


    def compile(self, shader_includes=True):
        if shader_includes:
            self._shader = Panda3dShader.make(self.language, do_shader_includes(self.vertex), do_shader_includes(self.fragment), do_shader_includes(self.geometry))
        else:
            self._shader = Panda3dShader.make(self.language, self.vertex, self.fragment, self.geometry)
        self.compiled = True


    @classmethod
    def load(cls, language=Panda3dShader.SL_GLSL, vertex=None, fragment=None, geometry=None, **kwargs):
        parts = {"vertex": vertex, "fragment": fragment, "geometry": geometry}
        parts = {k: v for k, v in parts.items() if v}

        folders = (  # folder search order
            application.asset_folder,
        )

        for sh, name in parts.items():
            for folder in folders:
                for filename in folder.glob('**/' + name):
                    with filename.open("rt") as f:
                        parts[sh] = f.read()

        parts.update(kwargs)
        return cls(language, **parts)


if __name__ == '__main__':
    from time import perf_counter
    t = perf_counter()
    from ursina import *
    from ursina import Ursina, Entity, held_keys, scene, EditorCamera

    app = Ursina()
    Entity(model='cube', shader=Shader())
    EditorCamera()
    print('ttttttttttttt', perf_counter() - t)
    def input(key):
        if held_keys['control'] and key == 'r':
            reload_shaders()

    def reload_shaders():
        for e in scene.entities:
            if hasattr(e, '_shader'):
                print('-------', e.shader)
                # e._shader = Panda3dShader.make(language, vertex, fragment, geometry)

    app.run()



from ursina import *


app = Ursina()

steps = [
'''
a




''',

]

t = Text(font='VeraMono.ttf', origin=(-.5,.5))




app.run()


'''
installation
basics / overview
cheat sheet
tutorials
example projects
faq
how to create a build

'''
The Entity god class

Most things are entities. The camera, a Button, Text, they all inherit from Entity.
This class



The Camera




Working with the Button class

Creating a button
b = Button()

Transforming the button
Since Button inherits Entity, we can transform in the same way:
b.position = (-.5, 0)
b.scale = .2

The 'origin' attribute can be very useful for ui elements.
By default it's in the center of the button, (0,0):
-------
|  *  |
-------

b.origin = (-.5, .5) # move the origin to the top left
*------
|     |
-------





There multiple way to define what happen when we click on a Button.
In any case we'd have to set the on_click attribute.

def button_click():
    print('click')

b.on_click = button_click # assign a method

b.on_click = Func(print, 'click') # assign a Func. This will call print with the argument 'click'

b.on_click = Sequence(Wait(1), Func(print, 'click delayed'))



from ursina import *

app = Ursina(forced_aspect_ratio=16/9, editor_ui_enabled=False)
window.color = color.black
camera.fov = 90
text = Text('press space to start', origin=[0,0])

outer_shape = Entity(model=Cylinder(16, mode='line', thickness=1), color=color.azure, scale=80, scale_y=16, x=-20, y=-5, enabled=False)
inner_shape = Entity(parent=outer_shape, model=Cylinder(16, mode='line', thickness=1), scale=.25, scale_y=1, color=outer_shape.color)

playing = False

def input(key):
    global playing

    if not playing and key == 'space':
        playing = True
        text.enabled = False
        outer_shape.enabled = True

def update():
    if playing:
        outer_shape.rotation_y += 1
        outer_shape.rotation_x -= .05

app.run()



from ursina import *
from tinyblend import BlenderFile


def load_level(name, path=application.asset_folder):
    if not name.endswith('.blend'):
        name += '.blend'

    files = list(path.glob(f'**/{name}'))
    print('files:', files)
    if not files:
        print(f'error: level {name} not found')
        return

    blend_file = files[0]
    # import bpy
    # dir(bpy.data)
    #
    with open(blend_file, 'rb') as f:
        blender_version_number = (f.read(12).decode("utf-8"))[-3:]   # get version from start of .blend file e.g. 'BLENDER-v280'
        blender_version_number = blender_version_number[0] + '.' + blender_version_number[1:2]
        print('blender_version:', blender_version_number)
        if blender_version_number in application.blender_paths:
            blender = application.blender_paths[blender_version_number]
        else:
            print('using default blender version')
            blender = application.blender_paths['default']

    a = subprocess.check_output(f'''"{blender}" "{blend_file}" --background --python {application.asset_folder/'blender_scene_parsing_script.py'}''', shell=True)
    a = a.decode( "utf-8")
    print('a:', a)
    # blend = BlenderFile(file)
    # structs = blend.list('Object')
    # # structs = blend.data
    # for o in structs:
    #     # print(e.collection)
    #     object_name = o.id.name.decode( "utf-8").replace(".", "_")[2:]
    #     object_name = object_name.split('\0', 1)[0]
    #     print('name:', object_name)

if __name__ == '__main__':
    load_level('blender_level_editor_test_scene')



from ursina import *


@generate_properties_for_class()
class Slider(Entity):
    def __init__(self, min=0, max=1, default=None, height=Text.size, text='', dynamic=False, radius=Text.size/2, bar_color=color.black66, **kwargs):
        super().__init__(add_to_scene_entities=False) # add later, when __init__ is done
        self.parent = camera.ui
        self.vertical = False
        self.min = min
        self.max = max

        if default is None:
            default = min
        self.default = default
        self.step = 0                   # for example, setting .step to 1, would make the slider snap to the closest integer.
        self.height = height

        self.on_value_changed = None    # set this to a function you want to be called when the slider changes
        self.setattr = None             # set this to (object, 'attrname') to set that value when the slider changes

        self.label = Text(parent=self, origin=(0.5, 0), x=-0.025, text=text)

        self.bg = Entity(parent=self, model=Quad(scale=(.525, height), radius=radius, segments=3),
            origin_x=-0.25, collider='box', color=bar_color)

        self.knob = Draggable(parent=self, min_x=0, max_x=.5, min_y=0, max_y=.5, step=self.step,
            model=Quad(radius=Text.size/2, scale=(Text.size, height)), collider='box', color=color.light_gray,
            text='0', text_origin=(0, -.55), z=-.1)

        def bg_click():
            self.knob.x = mouse.point[0]
            self.knob.start_dragging()
        self.bg.on_click = bg_click

        def drop():
            self.knob.z = -.1
            if self.setattr:
                if isinstance(self.setattr[0], dict):   # set value of dict
                    self.setattr[0][self.setattr[1]] = self.value
                else:                                   # set value of Entity
                    setattr(self.setattr[0], self.setattr[1], self.value)

            if self.on_value_changed:
                self.on_value_changed()


        self.knob.drop = drop
        self._prev_value = self.default
        self.value = self.default
        self.dynamic = dynamic    # if set to True, will call on_value_changed() while dragging. if set to False, will only call on_value_changed() after dragging.


        self.knob.text_entity.text = str(round(self.default, 2))

        for key, value in kwargs.items():
            setattr(self, key, value)

        if self.vertical:
            self.rotation_z = -90
            self.knob.lock = (1,0,0)
            self.knob.text_entity.rotation_z = 90
            self.knob.text_entity.position = (.015,0)
        else:
            self.knob.lock = (0,1,1)
            self.knob.text_entity.y = height/2

        scene.entities.append(self)


    def value_getter(self):
        val = lerp(self.min, self.max, self.knob.x * 2)
        if isinstance(self.step, int) and not self.step == 0:
            val = int(round(val, 0))

        return val

    def value_setter(self, value, call_on_value_changed=True):
        self.knob.x = (value - self.min) / (self.max - self.min) / 2
        self.slide(call_on_value_changed=call_on_value_changed)


    def step_setter(self, value):
        self._step = value
        self.knob.step = value / (self.max-self.min) / 2

    def update(self):
        if self.knob.dragging:
            self.slide()

    def slide(self, call_on_value_changed=True):
        t = self.knob.x / .5

        if self.step > 0:
            if isinstance(self.step, int) or self.step.is_integer():
                self.knob.text_entity.text = str(self.value)

        if call_on_value_changed and self.dynamic and self._prev_value != t:
            if self.on_value_changed:
                self.on_value_changed()

            if self.setattr:
                target_object, attr = self.setattr
                setattr(target_object, attr, self.value)

            self._prev_value = t

        invoke(self._update_text, delay=1/60)

    def _update_text(self):
            self.knob.text_entity.text = str(round(self.value, 2))


    def __setattr__(self, name, value):
        if name == 'eternal':
            try:
                self.label.eternal = value
                self.bg.eternal = value
                self.knob.eternal = value
            except:
                pass
        try:
            super().__setattr__(name, value)
        except Exception as e:
            return e



class ThinSlider(Slider):
    def __init__(self, *args, **kwargs):
        kwargs['height'] = Text.size
        super().__init__(*args, **kwargs)
        self.bg.model = Quad(scale=(.525, Text.size/5), radius=Text.size/10, segments=3)
        self.bg.origin = self.bg.origin
        self.bg.color = color.text_color
        self.bg.highlight_color = color.text_color
        self.knob.color = lerp(color.text_color, color.inverse(color.text_color), .1)
        self.label.color = color.text_color




if __name__ == '__main__':
    app = Ursina()

    box = Entity(model='cube', origin_y=-.5, scale=1, color=color.orange)

    def scale_box():
        box.scale_y = slider.value
        print(thin_slider.value)

    slider = Slider(0, 20, default=10, height=Text.size*3, y=-.4, step=1, on_value_changed=scale_box, vertical=True)

    thin_slider = ThinSlider(text='height', dynamic=True, on_value_changed=scale_box)

    thin_slider.label.origin = (0,0)
    thin_slider.label.position = (.25, -.1)

    app.run()



import sys
sys.path.append("..")
from ursina import *
import subprocess

class OpenInFileExplorer():

    # def __init__(self):
    #     self.path = 'afafaf'

    def input(self, key):
        if key == 'right mouse down' and self.entity.hovered:
            print('yoooooo', os.name, 'explorer "' + self.path + '"')
            if os.name == 'nt':
                subprocess.Popen('explorer "' + self.path + '"')



from ursina import *


app = Ursina()

level = '''
###....#
#....###
##.....#
##.....#
##.....#
.#####.#
'''.strip()

grid = list(reversed(level.split('\n'))) # flip it vertically so the last line is the first element

window.color = color.black
camera.orthographic = True
camera.fov = 10
camera.position = (4,2)

level_geometry = Entity(model=Mesh(mode='point', thickness=.5), texture='circle')
for y, line in enumerate(grid):
    for x, char in enumerate(line):
        if char == '#':
            level_geometry.model.vertices.append(Vec3(x,y,0))

level_geometry.model.generate()
level_geometry.model.set_render_mode_perspective(True)

player = Button(parent=scene, text='@', text_color=color.lime, position=(4,2,-.1))


def input(key):
    if key in ('a', 'a hold'):
        walk((-1,0))
    if key in ('d', 'd hold'):
        walk((1,0))
    if key in ('w', 'w hold'):
        walk((0,1))
    if key in ('s', 's hold'):
        walk((0,-1))


def walk(direction):
    target_x = int(player.x + direction[0])
    target_y = int(player.y + direction[1])

    collided = grid[target_y][target_x] in ('#', '|', '/', '\\', 'A')

    if not collided:
        player.position += direction


app.run()



Altering a 3d model
To change a model, just change it's attributes like .vertices, .triangles, .colors, .uvs, .normals, or .mode.
e = Entity(model='cube')

# move vertices above the origin up 1, to make a capsule
for v in e.model.vertices:
    if v.y > 0:
        v.y += 1

e.model.generate() # call this to update the mesh

You can also create a Mesh by scratch if you want that.
If you set model.mode = 'line' and call model.generate() it will render it as lines, but
keep in mind it might not look exactly as expected it if it's mean to be solid.

To set window display_mode
window.display_mode = 'wireframe' # render everything as wireframe

You can also press F10 to toggle between default, wireframe, colliders, normals. F9 to return to default.


UV Scrolling
Modifying a Mesh instance and regenerating it every frame is not the best way to make texture scrolling, however.
It's too expensive to do that every frame so you can set 'texture_offset' and 'texture_scale' instead. I noticed they're not
listed in the API reference and had actully forgotten about them myself. :P


Lighting
Lights are not really supported at the moment. The reflectivity is a work in progress as well, since I couldn't get it work like I wanted.
You can however take a look at panda3d's documentation and use the lights from there if you'd like.
Since ursina is made on top of panda3d you can always fall back to that control things at a lower level.

There's a really basic shader called 'basic_lighting' you can use of you just want to see the shape of the model better. No cast shadows included, but at least it's not flat shaded.
Or you could set window.display_mode = 'normals'. Keep in mind the model needs normals for these to work.
If the model doesn't have them you can generate them with entity.model.generate_normals(). It can take whole to do that though, especially for bigger meshes.

Often in games it's a good idea to bake the shadows anyway, as real-time shadows tend to multiply the draw calls.



from panda3d.core import Shader


basic_lighting_shader = Shader.make(Shader.SL_GLSL,
vertex='''
#version 130
uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelMatrixInverseTranspose;
uniform mat4 transform_matrix;
in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
in vec3 p3d_Normal;
out vec2 texcoord;
out vec3 world_space_normal;

void main() {
  gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
  texcoord = p3d_MultiTexCoord0;

  vec4 invcamx = transform_matrix[0];
  vec4 invcamy = transform_matrix[1];
  vec4 invcamz = transform_matrix[2];
  vec4 invcamw = transform_matrix[3];

  mat3 invcam = mat3(
                          invcamx[0], invcamx[1], invcamx[2],
                          invcamy[0], invcamy[1], invcamy[2],
                          invcamz[0], invcamz[1], invcamz[2]
                      );

  world_space_normal = normalize(p3d_Normal * invcam);
}
''',

fragment='''
#version 130

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;
in vec2 texcoord;
in vec3 world_space_normal;

uniform vec4 top_color;
uniform vec4 bottom_color;

out vec4 fragColor;


void main() {
    vec4 norm = vec4(world_space_normal*0.5+0.5, 1);
    // float grey = 0.21 * norm.r + 0.71 * norm.g + 0.07 * norm.b;
    // norm = vec4(grey, grey, grey, 1);
    vec4 color = texture(p3d_Texture0, texcoord) * p3d_ColorScale;
    // vec4 color = vec4(1.0, 0.0, 0.0, 0.0);
    float normal_y = world_space_normal.y + .5;
    color.r = mix(bottom_color.r, top_color.r, normal_y);
    // vec4 color = bottom_color;

    fragColor = color.rgba;
}


''', geometry='')


if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import *
    app = Ursina()
    window.color=color.black

    # e = Entity(model='sphere', shader=basic_lighting_shader)
    # e.setShaderInput('transform_matrix', e.getNetTransform().getMat())
    shader = basic_lighting_shader

    a = WhiteCube(shader=shader)
    b = AzureSphere(shader=shader, rotation_y=180, x=3)
    # AzureSphere(shader=a.shader, y=2)
    GrayPlane(scale=10, y=-2, texture='shore')

    Sky(color=color.light_gray)
    EditorCamera(rotate_around_mouse_hit=False)

    for e in (a, b):
        e.set_shader_input('transform_matrix', e.getNetTransform().getMat())
        e.set_shader_input('top_color',  (255, 0, 0, 255))
        e.set_shader_input('bottom_color', color.red)


    def update():
        b.rotation_y += 1
        b.set_shader_input('transform_matrix', b.getNetTransform().getMat())

    EditorCamera()

    app.run()



from ursina import *


app = Ursina()

cube_colors = [
    color.pink,     # right
    color.orange,   # left
    color.white,    # top
    color.yellow,   # bottom
    color.azure,    # back
    color.green,    # front
]

# make a model with a separate color on each face
combine_parent = Entity(enabled=False)
for i, direction in enumerate((Vec3.right, Vec3.up, Vec3.forward)):

    e = Entity(parent=combine_parent, model='plane', origin_y=-.5, texture='white_cube', color=cube_colors[i*2])
    e.look_at(direction, Vec3.up)

    e_flipped = Entity(parent=combine_parent, model='plane', origin_y=-.5, texture='white_cube', color=cube_colors[(i*2)+1])
    e_flipped.look_at(-direction, Vec3.up)

combine_parent.combine()


# place 3x3x3 cubes
cubes = []
for x in range(3):
    for y in range(3):
        for z in range(3):
            e = Entity(model=copy(combine_parent.model), position=Vec3(x,y,z) - (Vec3(3,3,3)/3), texture='white_cube')
            cubes.append(e)


# rotate a side when we click on it
collider = Entity(model='cube', scale=3, collider='box', visible=False)

def collider_input(key):
    if mouse.hovered_entity == collider:
        if key == 'left mouse down':
            rotate_side(mouse.normal, 1)
        elif key == 'right mouse down':
            rotate_side(mouse.normal, -1)

collider.input = collider_input


rotation_helper = Entity()


def rotate_side(normal, direction=1, speed=1):
    if normal == Vec3(1,0,0):
        [setattr(e, 'world_parent', rotation_helper) for e in cubes if e.x > 0]
        rotation_helper.animate('rotation_x', 90 * direction, duration=.15*speed, curve=curve.linear, interrupt='finish')
    elif normal == Vec3(-1,0,0):
        [setattr(e, 'world_parent', rotation_helper) for e in cubes if e.x < 0]
        rotation_helper.animate('rotation_x', -90 * direction, duration=.15*speed, curve=curve.linear, interrupt='finish')

    elif normal == Vec3(0,1,0):
        [setattr(e, 'world_parent', rotation_helper) for e in cubes if e.y > 0]
        rotation_helper.animate('rotation_y', 90 * direction, duration=.15*speed, curve=curve.linear, interrupt='finish')
    elif normal == Vec3(0,-1,0):
        [setattr(e, 'world_parent', rotation_helper) for e in cubes if e.y < 0]
        rotation_helper.animate('rotation_y', -90 * direction, duration=.15*speed, curve=curve.linear, interrupt='finish')

    elif normal == Vec3(0,0,1):
        [setattr(e, 'world_parent', rotation_helper) for e in cubes if e.z > 0]
        rotation_helper.animate('rotation_z', -90 * direction, duration=.15*speed, curve=curve.linear, interrupt='finish')
    elif normal == Vec3(0,0,-1):
        [setattr(e, 'world_parent', rotation_helper) for e in cubes if e.z < 0]
        rotation_helper.animate('rotation_z', 90 * direction, duration=.15*speed, curve=curve.linear, interrupt='finish')


    invoke(reset_rotation_helper, delay=.2*speed)

    if speed:
        collider.ignore_input = True
        @after(.25*speed)
        def _():
            collider.ignore_input = False
            check_for_win()


def reset_rotation_helper():
    [setattr(e, 'world_parent', scene) for e in cubes]
    rotation_helper.rotation = (0,0,0)


win_text_entity = Text(y=.35, text='', color=color.green, origin=(0,0), scale=3)

def check_for_win():
    if {e.world_rotation for e in cubes} == {Vec3(0,0,0)}:
        win_text_entity.text = 'SOLVED!'
        win_text_entity.appear()
    else:
        win_text_entity.text = ''


def randomize():
    faces = (Vec3(1,0,0), Vec3(0,1,0), Vec3(0,0,1), Vec3(-1,0,0), Vec3(0,-1,0), Vec3(0,0,-1))
    for i in range(20):
        rotate_side(normal=random.choice(faces), direction=random.choice((-1,1)), speed=0)

randomize_button = Button(text='randomize', color=color.azure, position=(.7,-.4), on_click=randomize)
randomize_button.fit_to_text()

window.color = color._16
EditorCamera()

app.run()



from ursina import color, load_texture
from ursina.shader import Shader
from ursina.ursinastuff import Func
from ursina.vec2 import Vec2

projector_shader = Shader(name='projector_shader', language=Shader.GLSL, vertex = '''
#version 140
uniform mat4 p3d_ModelViewProjectionMatrix;
uniform mat4 p3d_ModelMatrix;

in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
uniform vec2 texture_scale;
uniform vec2 texture_offset;
out vec2 texcoord;
in vec4 p3d_Color;
out vec4 vertex_color;
uniform vec2 projector_uv_scale;
uniform vec2 projector_uv_offset;

out vec2 world_uv;
uniform float time;

out float z;

void main() {
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
    texcoord = (p3d_MultiTexCoord0 * texture_scale) + texture_offset;
    vertex_color = p3d_Color;

    world_uv = (p3d_ModelMatrix * p3d_Vertex).xz * projector_uv_scale;
    world_uv -= projector_uv_offset;
    world_uv += vec2(.5);
    z = gl_Position.z;
}
''',

fragment='''
#version 140

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;
in vec2 texcoord;
out vec4 fragColor;
in vec4 vertex_color;

uniform sampler2D projector_texture;
uniform vec4 projector_color;
in vec2 world_uv;

uniform struct p3d_FogParameters {
  vec4 color;
  float density;
  float start;
  float end;
  float scale; // 1.0 / (end - start)
} p3d_Fog;

in float z;

float inverse_lerp(float a, float b, float v) {
    return (v-a) / (b-a);
}

void main() {
    vec4 color = texture(p3d_Texture0, texcoord) * p3d_ColorScale;
    color.rgb -= texture(projector_texture, world_uv).r * (vec3(1.) - projector_color.rgb) * projector_color.a;
    color.rgb = mix(color.rgb, p3d_Fog.color.rgb, clamp(inverse_lerp(p3d_Fog.start, p3d_Fog.end, z), 0.0, 1.0)); // linear fog

    fragColor = color.rgba * vertex_color;
}''',

default_input = {
    'texture_scale': Vec2(1,1),
    'texture_offset': Vec2(0,0),
    'projector_texture' : Func(load_texture, 'vignette'),
    'projector_uv_scale' : Vec2(.05, .05),
    'projector_uv_offset' : Vec2(.0, .0),
    'projector_color' : color.black,
    'time' : 0.0,
}
)


if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    Entity.default_shader = projector_shader

    editor_camera = EditorCamera(rotation_x=30,)
    light = Entity(model='sphere', unlit=True)
    ground = Entity(model='plane', collider='box', scale=64, texture='grass', texture_scale=(4,4))

    random.seed(0)
    for i in range(16):
        Entity(model='cube', origin_y=-.5, scale=2, texture='brick', texture_scale=(1,2),
            x=random.uniform(-8,8),
            z=random.uniform(-8,8) + 8,
            collider='box',
            scale_y = random.uniform(2,3),
            color=color.hsv(0, 0, random.uniform(.9, 1))
        )

    scene.fog_density = (10, 200)

    projector_texture = load_texture('vignette', application.internal_textures_folder)
    projector_texture.repeat = False
    projector_shader.default_input['projector_texture'] = projector_texture

    def update():
        light.x += (held_keys['d'] - held_keys['a']) * time.dt * 3
        light.z += (held_keys['w'] - held_keys['s']) * time.dt * 3

        for e in scene.entities:
            if hasattr(e,'shader') and e.shader == projector_shader:
                e.set_shader_input('projector_uv_offset', light.world_position.xz * projector_shader.default_input['projector_uv_scale'])

    app.run()



from ursina import *
from ursina.prefabs.text_field import TextField


app = Ursina()
# app = Ursina(forced_aspect_ratio=16/9)

window.position=(0,0)
window.size = Vec2(1080,1920)*.5

window.color = color.hex('#282c34')
Text.default_font = 'VeraMono.ttf'
text_editor = TextField(register_mouse_input=True, line_height=1.3, x=-.25, y=.25, active=False, ignore=True)
text_editor.line_numbers.enabled=True
text_editor.bg.color = window.color

text_editor.text_entity.text_colors['default'] = color.hex('#abb2bf')
text_editor.text_entity.text_colors['cyan'] = color.hex('#61afef')
text_editor.text_entity.text_colors['purple'] = color.hex('#c678dd')
text_editor.text_entity.text_colors['orange'] = color.hex('#d19a66')
text_editor.text_entity.text_colors['string_color'] = color.hex('#98c379')
text_editor.text_entity.text_colors['red'] = color.hex('#e06c75')
text_editor.text_entity.text_colors['gray'] = color.hex('#5c6370')
text_editor.line_numbers.color = lerp(color.hex('#5c6370'), window.color, .5)

for e in range(10):
    text_editor.replacements[f' {e}'] = f' ☾orange☽{e}' # numbers
    text_editor.replacements[f'={e}'] = f'=☾orange☽{e}' # numbers
    text_editor.replacements[f'.{e}'] = f'☾orange☽.{e}' # numbers
    text_editor.replacements[f'-{e}'] = f'☾orange☽-{e}' # numbers
    text_editor.replacements[f'({e}'] = f'(☾orange☽{e}' # numbers


text_editor.replacements.update({
    **{f'{e}': f'☾cyan☽{e}' for e in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'},

    'from ':    f'☾purple☽from ☾default☽',
    'import ':  f'☾purple☽import ☾default☽',
    'def ':     f'☾purple☽def ☾default☽',
    'for ':     f'☾purple☽for ☾default☽',
    'if ':      f'☾purple☽if ☾default☽',
    'with ':    f'☾purple☽with ☾default☽',
    ' in ':     f'☾purple☽ in ☾default☽',

    '\',':    f'\'☾default☽,',   # end quote
    '\':':    f'\':☾default☽',   # end quote
    '\')':    f'\')☾default☽',   # end quote
    '\']':    f'\'☾default☽]',   # end quote
    ' \'':    f' ☾string_color☽\'', # start quote
    '=\'':    f'=☾string_color☽\'', # start quote
    '(\'':    f'(☾string_color☽\'', # start quote
    '[\'':    f'[☾string_color☽\'', # start quote

    '=':        f'☾purple☽=☾default☽',
    '+':        f'☾purple☽+☾default☽',
    '-':        f'☾purple☽-☾default☽',
    '*':        f'☾purple☽*☾default☽',

    'print(':   f'☾func_color☽print☾default☽(',
    'range(':   f'☾func_color☽range☾default☽(',
    '__init__': f'☾func_color☽__init__☾default☽',
    'super':    f'☾func_color☽super☾default☽',

    'update(':   f'☾cyan☽update☾default☽(',
    'input(':    f'☾cyan☽input☾default☽(',

    'class ':   f'☾class_color☽class ☾default☽',
    'self.':    f'☾class_color☽self☾default☽.',
    '(self)':   f'(☾class_color☽self☾default☽)',
    'self,':    f'☾class_color☽self☾default☽,',

    '.':    f'.☾red☽',
    ' ':    f' ☾default☽',
    '(':    f'☾default☽(',
    ')':    f'☾default☽)',
    ', ':    f'☾default☽, ',
    })

code_sections = []
comment_sections = []

with open('how_to_move_forward.py', encoding='utf8') as f:
# with open('rotate_camera_around_point.py', encoding='utf8') as f:
    lines = f.read().split('\n')
    lines = [e for e in lines if not e.strip().endswith('#!hide') if e]

    mode = 'code'
    text = ''

    for i, l in enumerate(lines):
        if mode == 'code' and l.strip().startswith('#'):
            if text:
                code_sections.append(text)
            text = ''
            mode = 'comment'

        elif mode == 'comment' and not l.strip().startswith('#'):
            if text:
                comment_sections.append(text)
            text = ''
            mode = 'code'

        text += l + '\n'

    if mode == 'comment' and text:
        comment_sections.append(text)
    if mode == 'code' and text:
        code_sections.append(text)


parts = list(zip(comment_sections, code_sections))

i = 0
seq = None
def input(key):
    if key == 'space':
        show_next()


def show_next():
    global i, seq
    if i >= len(parts):
        return

    text_editor.render()

    if seq:
        seq.finish()

    text_editor.animate_y(.05+ text_editor.text.count('\n')*.025*text_editor.line_height, duration=1, curve=curve.in_out_cubic)
    def add_text(char):
        text_editor.text += char
        text_editor.render()

    seq = Sequence(
        Wait(1),
        # Func(add_text, parts[i][0]+'\n'),
        # Wait(len(parts[i][0])*.0125),
        )
    for char in parts[i][0]:
        seq.append(Func(add_text, char))
        seq.append(.025)
    seq.append(Func(add_text, '\n'))
    seq.append(Wait(.5))

    for char in parts[i][1]:
        seq.append(Func(add_text, char))
        seq.append(.05)

    seq.append(Func(add_text, '\n'))
    seq.append(Wait(.5))
    seq.append(Func(show_next))
    seq.start()

    i += 1
    print(i)

app.run()



import sys
sys.path.append("..")
from ursina import *

class SelectionButton():

    left_shift = False

    def __init__(self):
        self.entity = None
        self.selection_target = None
        self.dragging = False

    def input(self, key):
        if key == 'left shift':
            left_shift = True
        if key == 'left shift up':
            left_shift = False


        if key == 'left mouse down' and self.entity.hovered:
            # if not left_shift:
            scene.editor.selection.clear()

            for b in scene.editor.hierarchy_panel.buttons:
                b.color = color.panda_button

            self.entity.color = color.blue
            self.dragging = True

            scene.editor.selection.append(self.selection_target)





        if key == 'left mouse up':
            if self.dragging and mouse.hovered_entity is not self.entity and mouse.hovered_entity.selection_button:
                print('reparent to:', mouse.hovered_entity.selection_button.selection_target.name)
                self.selection_target.reparent_to(mouse.hovered_entity.selection_button.selection_target)
                scene.editor.hierarchy_panel.populate()
            self.dragging = False

    # def update(self):
    #     if self.dragging:
    #         # print('updateing slection button')
    #         print(self.entity.scale[1])
    #         print(mouse.delta[1])
    #         if mouse.delta[1] > self.entity.scale[1]:
    #             pass



from ursina import *
from ursina import color as ursinacolor


class ExplosiveSceneTransition(Entity):
    def __init__(self, color=ursinacolor.black, **kwargs):
        super().__init__()
        self.world_parent = camera

        for i in range(10):
            e = Entity(
                parent = self,
                # model = Cone(direction=(0,0,1)),
                model = 'sphere',
                scale = random.uniform(.5, 2),
                scale_z = 20,
                origin_z = -.5,
                color = ursinacolor.clear,
                z = -1
                )
            e.look_at((
                random.uniform(-1, 1),
                random.uniform(-1, 1),
                random.uniform(-1, 1)
                ))
            e.animate_scale_z(0, .3, delay=random.uniform(0,.2))
            e.animate_color(color, delay=.3)
            e.fade_out(duration=.5)
            window.overlay.fade_in(.4, delay=.4)
            # e.animate_scale((0,0,0), .3, delay=random.uniform(.2,.5))
        for i in range(3):
            c = Entity(
                parent = self,
                model = Circle(24),
                scale = 0,
                color = color,
                z = -5
                )
            c.animate_scale(Vec3(i+1,i+1,i+1)*6, duration=i/10 +.1, delay=.5, curve=curve.in_bounce)

        # window.overlay.fade_out(delay=1)
        destroy(self, delay=1)

if __name__ == '__main__':
    app = Ursina()
    random.seed(0)
    ExplosiveSceneTransition(color=color.black)
    app.run()



if __name__ == '__main__':
    from ursina import *


    app = Ursina()

    face_parent = Entity(texture='sliced_cube_template')
    # middle = Entity(parent=face_parent, model='quad', scale=.5, texture='sliced_cube_template', texture_scale=Vec2(.5), texture_offset=Vec2(.25))
    #
    # for corner in [e/2 for e in [Vec2(1,1), Vec2(1,-1), Vec2(-1,-1), Vec2(-1,1)]]:
    #     print(corner)
    #     Entity(parent=face_parent, model='quad', origin=corner, position=corner, scale=.25, texture='sliced_cube_template', texture_scale=Vec2(.25), texture_offset=(corner+Vec2(.5))*.75)
    #
    # for side in [e/2 for e in [Vec2(0,1), Vec2(1,0), Vec2(0,-1), Vec2(-1,0)]]:
    #     Entity(parent=face_parent, model='quad', origin=side, position=side, scale=Vec2(.25,.25)+abs(side.yx), texture='sliced_cube_template', texture_scale=Vec2(.25,.8), texture_offset=(0,0), color=color.white)


    # face_parent.combine()
    quad = load_model('quad', use_deepcopy=True)

    vertices = []
    uvs = []

    vertices.extend([v/2 for v in quad.vertices])
    uvs.extend([(uv/2)+Vec2(1/4) for uv in quad.uvs])

    for corner in [e/2 for e in [Vec2(1,1), Vec2(1,-1), Vec2(-1,-1), Vec2(-1,1)]]:
        vertices.extend([(v/4) + corner*.75 for v in quad.vertices])
        uvs.extend([(uv/4)+(corner+Vec2(.5))*.75 for uv in quad.uvs])

    # for side in [e/2 for e in [Vec2(0,1), Vec2(1,0), Vec2(0,-1), Vec2(-1,0)]]:
    #     vertices.extend([(v*Vec3(*abs(side.yx))) + side.yx + Vec3(0,0,-.5)for v in quad.vertices])
    #     uvs.extend([(uv/4)+(side+Vec2(.5))*.75 for uv in quad.uvs])
    #     # tris.append([i for i in range(len(vertices), len(vertices)+4)])
    #     # Entity(parent=face_parent, model='quad', position=side*.75, scale=Vec2(.25,.25)+abs(side.yx), texture='sliced_cube_template', texture_scale=Vec2(.25,.8), texture_offset=(0,.1), color=color.white)

    for side, uv_offset in zip([e/2 for e in [Vec2(0,1), Vec2(1,0), Vec2(0,-1), Vec2(-1,0)]], [Vec2(.25,0), Vec2(.75,.25), Vec2(.25,.75), Vec2(0,.25)]):
        size = Vec3(.25,.25, 1) + (abs(side.yx)*.5)
        position = (Vec3(*side,0) * .75)

        vertices.extend([(v * size) + position for v in quad.vertices])
        uvs.extend([(uv*size) + uv_offset for uv in quad.uvs])

    m = Mesh(vertices=vertices,  uvs=uvs)
    Entity(model=m, texture='sliced_cube_template')


    Entity(model='wireframe_quad', color=color.green, alpha=.5)
    # Entity(model=m, texture='shore')
    EditorCamera()
    app.run()



fragment_code = fragment='''
#version 130
 precision highp float;

 out vec4 fragColor;

 uniform vec3 eye;
 in vec3 initialRay;
 uniform float textureWeight;
 uniform float timeSinceStart;
 uniform sampler2D p3d_Texture0;
 uniform float glossiness;
 vec3 roomCubeMin = vec3(-4.99995, 0, -4.99995);
 vec3 roomCubeMax = vec3(4.99995, 9.9999, 4.99995);uniform vec3 light;
 uniform vec3 sphereCenter0;
 uniform float sphereRadius0;

 vec2 intersectCube(vec3 origin, vec3 ray, vec3 cubeMin, vec3 cubeMax) {
   vec3 tMin = (cubeMin - origin) / ray;
   vec3 tMax = (cubeMax - origin) / ray;
   vec3 t1 = min(tMin, tMax);
   vec3 t2 = max(tMin, tMax);
   float tNear = max(max(t1.x, t1.y), t1.z);
   float tFar = min(min(t2.x, t2.y), t2.z);
   return vec2(tNear, tFar);
 }

 vec3 normalForCube(vec3 hit, vec3 cubeMin, vec3 cubeMax) {
   if(hit.x < cubeMin.x + 0.0001) return vec3(-1.0, 0.0, 0.0);
   else if(hit.x > cubeMax.x - 0.0001) return vec3(1.0, 0.0, 0.0);
   else if(hit.y < cubeMin.y + 0.0001) return vec3(0.0, -1.0, 0.0);
   else if(hit.y > cubeMax.y - 0.0001) return vec3(0.0, 1.0, 0.0);
   else if(hit.z < cubeMin.z + 0.0001) return vec3(0.0, 0.0, -1.0);
   else return vec3(0.0, 0.0, 1.0);
 }

 float intersectSphere(vec3 origin, vec3 ray, vec3 sphereCenter, float sphereRadius) {
   vec3 toSphere = origin - sphereCenter;
   float a = dot(ray, ray);
   float b = 2.0 * dot(toSphere, ray);
   float c = dot(toSphere, toSphere) - sphereRadius*sphereRadius;
   float discriminant = b*b - 4.0*a*c;
   if(discriminant > 0.0) {
     float t = (-b - sqrt(discriminant)) / (2.0 * a);
     if(t > 0.0) return t;
   }
   return 10000.0;
 }

 vec3 normalForSphere(vec3 hit, vec3 sphereCenter, float sphereRadius) {
   return (hit - sphereCenter) / sphereRadius;
 }

 float random(vec3 scale, float seed) {
   return fract(sin(dot(gl_FragCoord.xyz + seed, scale)) * 43758.5453 + seed);
 }

 vec3 cosineWeightedDirection(float seed, vec3 normal) {
   float u = random(vec3(12.9898, 78.233, 151.7182), seed);
   float v = random(vec3(63.7264, 10.873, 623.6736), seed);
   float r = sqrt(u);
   float angle = 6.283185307179586 * v;
   vec3 sdir, tdir;
   if (abs(normal.x)<.5) {
     sdir = cross(normal, vec3(1,0,0));
   } else {
     sdir = cross(normal, vec3(0,1,0));
   }
   tdir = cross(normal, sdir);
   return r*cos(angle)*sdir + r*sin(angle)*tdir + sqrt(1.-u)*normal;
 }

 vec3 uniformlyRandomDirection(float seed) {
   float u = random(vec3(12.9898, 78.233, 151.7182), seed);
   float v = random(vec3(63.7264, 10.873, 623.6736), seed);
   float z = 1.0 - 2.0 * u;
   float r = sqrt(1.0 - z * z);
   float angle = 6.283185307179586 * v;
   return vec3(r * cos(angle), r * sin(angle), z);
 }

 vec3 uniformlyRandomVector(float seed) {
   return uniformlyRandomDirection(seed) * sqrt(random(vec3(36.7539, 50.3658, 306.2759), seed));
 }

 float shadow(vec3 origin, vec3 ray) {
   float tSphere0 = intersectSphere(origin, ray, sphereCenter0, sphereRadius0);
   if(tSphere0 < 1.0)
     return 0.0;
   return 1.0;
 }

vec3 calculateColor(vec3 origin, vec3 ray, vec3 light) {
    vec3 colorMask = vec3(1.0);
    vec3 accumulatedColor = vec3(0.0);
    for(int bounce = 0; bounce < 5; bounce++) {
        vec2 tRoom = intersectCube(origin, ray, roomCubeMin, roomCubeMax);
        float tSphere0 = intersectSphere(origin, ray, sphereCenter0, sphereRadius0);
        vec3 surfaceColor = vec3(0.8, 0.8, 0.8);
        float t = 10000.0;

        if(tRoom.x < tRoom.y)
            t = tRoom.y;

        if(tSphere0 < t) {
            t = tSphere0;
            surfaceColor = vec3(0.0, 1.0, 0.1);
        }

        vec3 hit = origin + ray * t;

        float specularHighlight = 0.0;
        vec3 normal;

        if(t == tRoom.y) {
            normal = -normalForCube(hit, roomCubeMin, roomCubeMax);
        if(hit.x < -4.99895)
            surfaceColor = vec3(0.1, 0.1, 0.1);
        else if(hit.x > 4.99895)
            surfaceColor = vec3(0.1, 0.1, 0.1);
            ray = cosineWeightedDirection(timeSinceStart + float(bounce), normal);
        }
        else if(t == 10000.0) {
            break;
        }
        else {
            if(false);

            else if(t == tSphere0)
                normal = normalForSphere(hit, sphereCenter0, sphereRadius0);

            ray = cosineWeightedDirection(timeSinceStart + float(bounce), normal);
        }
        vec3 toLight = light - hit;
        float diffuse = max(0.0, dot(normalize(toLight), normal));
        float shadowIntensity = shadow(hit + normal * 0.0001, toLight);
        colorMask *= surfaceColor;
        accumulatedColor += colorMask * (0.5 * diffuse * shadowIntensity);
        accumulatedColor += colorMask * specularHighlight * shadowIntensity;
        origin = hit;
    }
    return accumulatedColor;
}

 void main() {
   vec3 newLight = light + uniformlyRandomVector(timeSinceStart - 53.0) * 0.1;
   vec3 texture = texture2D(p3d_Texture0, gl_FragCoord.xy / vec2(512.0, 512)).rgb;
   fragColor = vec4(mix(calculateColor(eye, initialRay, newLight), texture, .5), 1.0);
   // fragColor = vec4(1,0,1, 1.0);
 }
'''
from panda3d.core import Shader

camera_contrast_shader = Shader.make(Shader.SL_GLSL,
vertex='''
#version 130
uniform mat4 p3d_ModelViewProjectionMatrix;
in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
out vec2 texcoord;

// attribute vec3 vertex;' +
uniform vec3 eye, ray00, ray01, ray10, ray11;
out vec3 initialRay;
void main() {
    vec2 percent = p3d_Vertex.xy * 0.5 + 0.5;
    initialRay = mix(mix(ray00, ray01, percent.y), mix(ray10, ray11, percent.y), percent.x);
    // gl_Position = p3d_Vertex;
    gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
    // texcoord = p3d_MultiTexCoord0;
}

// void main() {
//   gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
//   texcoord = p3d_MultiTexCoord0;
// }
''',

fragment=fragment_code, geometry='')


if __name__ == '__main__':
    from ursina import *
    app = Ursina()

    e = Entity(model='cube', color=color.orange)
    camera.shader = camera_contrast_shader
    # camera.set_shader_input('contrast', 1)

    eye = Vec3(0,0,0)

    t = 0

    mat = camera.lens.getProjectionMat()

    camera.set_shader_input('timeSinceStart', t)
    camera.set_shader_input('light', Vec3(0.4, 0.5, -0.6))
    camera.set_shader_input('eye', camera.position)
    camera.set_shader_input('ray00', mat[0][0])
    camera.set_shader_input('ray01', mat[0][1])
    camera.set_shader_input('ray10', mat[1][0])
    camera.set_shader_input('ray11', mat[1][1])

    camera.set_shader_input('sphereCenter0', Vec3(0,1,1))
    camera.set_shader_input('sphereRadius0', .25)

    sampleCount = 0
    textureWeight = sampleCount / (sampleCount + 1)

    def update():
      global t, sampleCount
      t += time.dt * .5
      # t = t % 1
      # // print(t)
      # camera.set_shader_input('timeSinceStart', t)
      camera.set_shader_input('timeSinceStart', time.time())

      sampleCount += 1
      textureWeight = sampleCount / (sampleCount + 1)
      camera.set_shader_input('eye', camera.position)
      camera.set_shader_input('sampleCount', sampleCount)
      camera.set_shader_input('textureWeight', textureWeight)


    EditorCamera()

    app.run()



from ursina import application
import time


Wait = float


class Func:
    def __init__(self, func, *args, **kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__ (self):
        return self.func(*self.args, **self.kwargs)


class Sequence:
    default_time_step = None

    def __init__(self, *args, unscaled=False, started=False, ignore_paused=False, loop=False, auto_destroy=False, entity=None, time_scale=1, name='sequence', **kwargs):
        super().__init__()
        self.args = list(args)
        self.t = 0
        self.time_step = Sequence.default_time_step
        self.time_scale = time_scale
        self.duration = 0
        self.funcs = []
        self.func_call_time = []
        self.func_finished_statuses = []
        self.unscaled = unscaled
        self.paused = False
        self.name = name
        self.entity = entity  # you can assign this to make the sequence pause when the entity is disabled or .ignore is True
        self.ignore_paused = ignore_paused
        self.loop = loop
        self.auto_destroy = auto_destroy
        self.started = started

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.generate()
        application.sequences.append(self)


    def generate(self):
        self.duration = 0
        self.funcs = []
        self.func_call_time = []
        self.func_finished_statuses = []

        for arg in self.args:
            if isinstance(arg, (int, float)):
                self.duration += arg

            elif callable(arg):
                self.funcs.append(arg)
                self.func_call_time.append(self.duration)
                self.func_finished_statuses.append(False)

        # print('-----------')
    def __str__(self):
        return '\n'.join([str(e) for e in zip(self.funcs, self.func_call_time, self.func_finished_statuses)])

    def __call__(self):
        self.start()
        return self


    def append(self, arg, regenerate=True):
        # print('---------------', arg, callable(arg))
        if not callable(arg) and not isinstance(arg, (int, float)):
            raise TypeError(f'Invalid type: {arg}. Must be callable, Func, Wait or float.')

        self.args.append(arg)
        if regenerate:
            self.generate()

    def extend(self, list):
        for e in list:
            self.append(e, regenerate=False)
        self.generate()

    def start(self):
        for i, f in enumerate(self.funcs):
            self.func_finished_statuses[i] = False

        self.t = 0
        self.started = True
        self.paused = False
        return self

    def pause(self):
        self.paused = True

    def resume(self):
        self.paused = False

    def finish(self):
        self.t = self.duration
        self.update()
        self.paused = False
        self.started = False

    def kill(self):
        if self in application.sequences:
            application.sequences.remove(self)
            del self

    @property
    def finished(self):
        return self.t >= self.duration


    def update(self):
        if not self.started:
            return

        if self.ignore_paused is False and (self.paused or application.paused):
            return

        if self.entity and (not self.entity.enabled or self.entity.ignore):
            return


        if self.time_step is None:
            if not self.unscaled:
                self.t += time.dt * self.time_scale
            else:
                self.t += time.dt_unscaled
        else:
            self.t += self.time_step

        for i, f in enumerate(self.funcs):
            if not self.func_finished_statuses[i] and self.func_call_time[i] <= self.t:
                f()
                self.func_finished_statuses[i] = True


        if self.t >= self.duration:
            if self.loop:
                for i, f in enumerate(self.funcs):
                    self.func_finished_statuses[i] = False

                if time.dt > self.duration: # if delta time is too big, set t to 0 so it doesn't get stuck, but allow desync.
                    self.t = 0
                else:
                    self.t -= self.duration
                return

            if self.auto_destroy and self in application.sequences:
                application.sequences.remove(self)
                del self



if __name__ == '__main__':
    from ursina import *
    from ursina import Ursina, Entity
    app = Ursina()
    e = Entity(model='quad')
    def some_func():
        print('some_func')

    s = Sequence(
        some_func,
        1,
        Func(print, 'one'),
        Func(e.fade_out, duration=1),
        Wait(1),
        loop=True
        )

    for i in range(8):
        s.append(Func(print, i))
        s.append(Wait(.2))

    print(s)

    def input(key):
        actions = {'s' : s.start, 'f' : s.finish, 'p' : s.pause, 'r' : s.resume}
        if key in actions:
            actions[key]()

    app.run()



from ursina import *

class LoadSceneButton():

    def __init__(self):
        self.path = None

    def input(self, key):
        if key == 'left mouse down' and self.entity.hovered:
            if self.path:
                self.load_scene()

    def load_scene(self):
        scene_name = os.path.basename(self.path).split('.')[0]
        load_scene(scene_name)
        scene.editor.hierarchy_panel.populate()

        for e in scene.entities:
            try:
                e.editor_collider = 'box'
                e.collider.stash()
            except:
                print('couldnt add editor collider to:', e)



# Start by importing ursina and creating a window.
from ursina import *
app = Ursina()

# ## Using the built in platformer controller
# A simple way to get stared is to use the built in platformer controller.
# It's pretty basic, so you might want to write your own at a later point.
# It is however a good starting point, so let's import it like this:
from ursina.prefabs.platformer_controller_2d import PlatformerController2d
player = PlatformerController2d(y=1, z=.01, scale_y=1, max_jumps=2)

# You can change settings like jump_height, walk_speed, and gravity.
# If you want to larn more about how it works you can read its code here:
# https://github.com/pokepetter/ursina/blob/master/ursina/prefabs/platformer_controller_2d.py
#
# If we try to play the game right now, you'll fall for all infinity, so let's add a ground:
ground = Entity(model='quad', scale_x=10, collider='box', color=color.black)

# ## Making a "level editor"
# Now, it works, but it's a pretty boring game, so let's make a more interesting level.
# There are many ways to go about making a level, but for this we'll make an image
# where we can simply draw the level and then generate a level based on that.
#
# #image platformer_tutorial_level.png
# ↑
# Make sure to save this image to same folder or below as your script.
#
# To generate the level we'll loop through all the pixels in the image above and do
# something based on the color of the pixel. If it's white, it's air, so we'll skip it.
# Now, we *could* create an Entity for each tile, but that's slower to render than one Entity with a custom model.
# To make the model, we'll use the Mesh class. You'll have to be somewhat familiar with
# what meshes are to do this, and know what vertices and uvs are. However, we'll just be
# copying the quad model and offset the vertices. If we wanted to use a tileset with different
# graphics for each tile, we'd scale and offset the uvs as well.

quad = load_model('quad', use_deepcopy=True)


level_parent = Entity(model=Mesh(vertices=[], uvs=[]), texture='white_cube')
def make_level(texture):
    # destroy every child of the level parent.
    # This doesn't do anything the first time the level is generated, but if we want to update it several times
    # this will ensure it doesn't just create a bunch of overlapping entities.
    [destroy(c) for c in level_parent.children]

    for y in range(texture.height):
        collider = None
        for x in range(texture.width):
            col = texture.get_pixel(x,y)

            # If it's black, it's solid, so we'll place a tile there.
            if col == color.black:
                level_parent.model.vertices += [Vec3(*e) + Vec3(x+.5,y+.5,0) for e in quad.generated_vertices] # copy the quad model, but offset it with Vec3(x+.5,y+.5,0)
                level_parent.model.uvs += quad.uvs
                # Entity(parent=level_parent, position=(x,y), model='cube', origin=(-.5,-.5), color=color.gray, texture='white_cube', visible=True)
                if not collider:
                    collider = Entity(parent=level_parent, position=(x,y), model='cube', origin=(-.5,-.5), collider='box', visible=False)
                else:
                    # instead of creating a new collider per tile, stretch the previous collider right.
                    collider.scale_x += 1
            else:
                collider = None

            # If it's green, we'll place the player there. Store this in player.start_position so we can reset the plater position later.
            if col == color.green:
                player.start_position = (x, y)
                player.position = player.start_position

    level_parent.model.generate()

make_level(load_texture('platformer_tutorial_level'))   # generate the level

# ## Positioning the camera
# Set the camera to orthographic so there's no perspective.
# Move the camera to the middle of the level and set the fov so the level fits nicely.
# Setting the fov on an orthographic camera means setting how many units vertically the camera can see.
camera.orthographic = True
camera.position = (30/2,8)
camera.fov = 16


# ## Adding player graphics and animations
# Loads an image sequence as a frame animation.
# So if you have some frames named image_000.png, image_001.png, image_002.png and so on,
# you can load it like this: Animation('image')
# You can also load a .gif by including the file type: Animation('image.gif')
# player.walk_animation = Animation('player_walk')

# the platformer controller has an Animator and will toggle the state based on
# whether it's standing still, is walking or is jumping.
# All the Animator does is to make sure only one Animation is enabled at the same time.
# Otherwise they would overlap.
# self.animator = Animator({'idle' : None, 'walk' : None, 'jump' : None})
player.traverse_target = level_parent
enemy = Entity(model='cube', collider='box', color=color.red, position=(16,5,-.1))
def update():
    if player.intersects(enemy).hit:
        print('die')
        player.position = player.start_position


# ## Start the game
EditorCamera()
app.run()

# ## Adding level graphics
#
# Coming later
pass



from ursina import *


class TerrainSystem(Entity):
    def __init__(self, height_values, subsection_dimensions=[64, 64], player=None, **kwargs):
        super().__init__(**kwargs)

        self.subsection_dimensions = subsection_dimensions
        w = len(height_values)
        h = len(height_values[0])
        self.subsections = [[None for y in range(h)] for x in range(w)]
        self.player = player

        for y in range(h // subsection_dimensions[1]):
            for x in range(w // subsection_dimensions[0]):
                print('-----', x,y)
                hv = [[None for y in range(64)] for x in range(64)]

                for __y in range(64):
                    for __x in range(64):
                        hv[__x][__y] = height_values[63*x + __x][63*y + __y]

                print(hv)
                tile = Entity(parent=self,
                    model=Terrain(height_values=hv),
                    x=x*63, z=y*63, scale=Vec3(64, 1, 64),
                    texture_scale=(1/4, 1/2),
                    texture_offset=(x/4, y/2),
                    texture='heightmap_1'
                    )


                self.subsections[x][y] = tile

    def update(self):
        if not self.player:
            return





if __name__ == '__main__':
    app = Ursina()
    from ursina.models.procedural.terrain import texture_to_height_values
    hv = texture_to_height_values('heightmap_1', skip=1)
    TerrainSystem(hv, scale_y=32)

    EditorCamera()
    app.run()
